# Copyright 2008 Divmod, Inc. See LICENSE file for details

from zope.interface import implements

from twisted.trial.unittest import TestCase
from twisted.trial.util import suppress as SUPPRESS
from twisted.python.usage import UsageError
from twisted.python.components import registerAdapter

from axiom.store import Store
from axiom.item import Item
from axiom.substore import SubStore
from axiom.userbase import LoginSystem
from axiom.attributes import boolean, integer, inmemory
from axiom.plugins.axiom_plugins import Create
from axiom.plugins.mantissacmd import Mantissa
from axiom.plugins.offeringcmd import SetFrontPage
from axiom.dependency import installOn

from axiom.test.util import CommandStubMixin

from nevow import rend, context, inevow
from nevow.inevow import IResource
from nevow.page import Element
from nevow.rend import NotFound
from nevow.flat import flatten
from nevow.tags import title, div, span, h1, h2
from nevow.testutil import FakeRequest

from xmantissa.ixmantissa import (
    IPublicPage, ITemplateNameResolver, INavigableElement, ISiteURLGenerator,
    IOfferingTechnician, INavigableFragment, ISiteRootPlugin, IWebViewer)
from xmantissa import signup
from xmantissa.website import APIKey, WebSite
from xmantissa.webapp import (PrivateApplication,
                              _AuthenticatedWebViewer)
from xmantissa.prefs import PreferenceAggregator
from xmantissa.port import SSLPort
from xmantissa.webnav import Tab
from xmantissa.offering import Offering, InstalledOffering, installOffering
from xmantissa.webtheme import theThemeCache
from xmantissa.sharing import shareItem, getEveryoneRole
from xmantissa.websharing import getDefaultShareID, UserIndexPage
from xmantissa.publicweb import (
    _AnonymousWebViewer, FrontPage, PublicAthenaLivePage,
    PublicNavAthenaLivePage, _PublicFrontPage, getLoader, AnonymousSite,
    _OfferingsFragment, _CustomizingResource, PublicPage, LoginPage)

from xmantissa.signup import PasswordResetResource
from xmantissa.test.test_offering import FakeOfferingTechnician
from xmantissa.test.test_websharing import TestAppPowerup, ITest
from xmantissa.test.test_webshell import WebViewerTestMixin
from xmantissa.test.test_website import SiteTestsMixin

class TestAppElement(Element):
    """
    View class for TestAppPowerup.
    """
    docFactory = object()       # masquerade as a valid Element, for the
                                # purposes of theme lookup.
    def __init__(self, original):
        self.original = original
        Element.__init__(self)

registerAdapter(TestAppElement, ITest, INavigableFragment)


class FakeTheme(object):
    """
    Trivial implementation of L{ITemplateNameResolver} which returns document
    factories from an in-memory dictionary.
    @ivar docFactories: C{dict} mapping fragment names to document factory
        objects.
    """
    def __init__(self, docFactories):
        self.docFactories = docFactories


    def getDocFactory(self, fragmentName, default=None):
        """
        Return the document factory for the given name, or the default value if
        the given name is unknown.
        """
        return self.docFactories.get(fragmentName, default)



class AnonymousWebViewerTests(WebViewerTestMixin, TestCase):
    """
    Tests for L{_AnonymousWebViewer}.
    """

    def setupPageFactory(self):
        """
        Create the page factory used by the tests.
        """
        self.pageFactory = _AnonymousWebViewer(self.siteStore)


    def test_roleIn(self):
        """
        L{_AnonymousWebViewer} should provide the Everyone role in the store it
        is asked about.
        """
        theRole = getEveryoneRole(self.adminStore)
        self.assertIdentical(
            self.pageFactory.roleIn(self.adminStore),
            theRole)



class FakeNavigableElement(Item):
    """
    Navigation contributing powerup tests can use to verify the behavior of the
    navigation renderers.
    """
    powerupInterfaces = (INavigableElement,)
    implements(*powerupInterfaces)

    dummy = integer()
    tabs = inmemory(
        doc="""
        The object which will be returned by L{getTabs}.
        """)

    def getTabs(self):
        """
        Return whatever tabs object has been set.
        """
        return self.tabs



class FakeTemplateNameResolver(object):
    """
    Template name resolver which knows about one template.

    @ivar correctName: The name of the template this resolver knows about.

    @ivar correctFactory: The template which will be returned for
        C{correctName}.
    """
    implements(ITemplateNameResolver)

    def __init__(self, correctName, correctFactory):
        self.correctName = correctName
        self.correctFactory = correctFactory


    def getDocFactory(self, name, default=None):
        """
        Return the default for all names other than C{self.correctName}.
        Return C{self.correctFactory} for that.
        """
        if name == self.correctName:
            return self.correctFactory
        return default



class FakePublicItem(Item):
    """
    Some item that is to be shared on an app store.
    """
    dummy = integer()



class FakeApplication(Item):
    """
    Fake implementation of an application installed by an offering.
    """
    implements(IPublicPage)

    index = boolean(doc="""
    Flag indicating whether this application wants to be included on the front
    page.
    """)



class TestHonorInstalledThemes(TestCase):
    """
    Various classes should be using template resolvers to determine which theme
    to use based on a site store.
    """
    def setUp(self):
        self.correctDocumentFactory = object()
        self.store = Store()
        self.fakeResolver = FakeTemplateNameResolver(
            None, self.correctDocumentFactory)

        def fakeConform(interface):
            if interface is ITemplateNameResolver:
                return self.fakeResolver
            return None
        self.store.__conform__ = fakeConform


    def test_offeringsFragmentLoader(self):
        """
        L{_OfferingsFragment.docFactory} is the I{front-page} template loaded
        from the store's ITemplateNameResolver.
        """
        self.fakeResolver.correctName = 'front-page'
        frontPage = FrontPage(store=self.store)
        offeringsFragment = _OfferingsFragment(frontPage)
        self.assertIdentical(
            offeringsFragment.docFactory, self.correctDocumentFactory)


    def test_loginPageLoader(self):
        """
        L{LoginPage.fragment} is the I{login} template loaded from the store's
        ITemplateNameResolver.
        """
        self.fakeResolver.correctName = 'login'
        page = LoginPage(self.store)
        self.assertIdentical(
            page.fragment, self.correctDocumentFactory)


    def test_passwordResetLoader(self):
        """
        L{PasswordResetResource.fragment} is the I{login} template loaded from
        the store's ITemplateNameResolver.
        """
        self.fakeResolver.correctName = 'reset'
        resetPage = PasswordResetResource(self.store)
        self.assertIdentical(
            resetPage.fragment, self.correctDocumentFactory)


class OfferingsFragmentTestCase(TestCase):
    """
    Tests for L{_OfferingsFragment}.
    """
    def setUp(self):
        """
        Set up stores and an offering.
        """
        store = Store(dbdir=self.mktemp())
        appStore1 = SubStore.createNew(store, ("app", "test1.axiom"))
        appStore2 = SubStore.createNew(store, ("app", "test2.axiom"))
        self.firstOffering = Offering(u'first offering', None, None, None, None,
                                      None, [])
        firstInstalledOffering = InstalledOffering(
            store=store, application=appStore1,
            offeringName=self.firstOffering.name)
        ss1 = appStore1.open()
        self.installApp(ss1)
        # (bypass Item.__setattr__)
        object.__setattr__(
            firstInstalledOffering, 'getOffering',
            lambda: self.firstOffering)

        secondOffering = Offering(u'second offering', None, None, None, None,
                                  None, [])
        secondInstalledOffering = InstalledOffering(
            store=store, application=appStore2,
            offeringName=secondOffering.name)
        # (bypass Item.__setattr__)
        object.__setattr__(secondInstalledOffering, 'getOffering',
                           lambda: secondOffering)

        self.fragment = _OfferingsFragment(FrontPage(store=store))


    def installApp(self, ss1):
        """
        Create a public item and share it as the default.
        """
        fpi = FakePublicItem(store=ss1)
        shareItem(fpi, toRole=getEveryoneRole(ss1),
                  shareID=getDefaultShareID(ss1))


    def test_offerings(self):
        """
        L{_OfferingsFragment.data_offerings} returns a generator of C{dict}
        mapping C{'name'} to the name of an installed offering with a
        shared item that requests a link on the default public page.
        """

        self.assertEqual(
            list(self.fragment.data_offerings(None, None)),
            [{'name': self.firstOffering.name}])



class OldOfferingsFragmentTestCase(OfferingsFragmentTestCase):
    """
    Test for deprecated behaviour of L{_OfferingsFragment}.
    """

    def installApp(self, ss1):
        """
        Install an L{IPublicPage} powerup.
        """
        fa = FakeApplication(store=ss1, index=True)
        ss1.powerUp(fa, IPublicPage)


    def test_offerings(self):
        """
        Test the deprecated case for rendering the offering list.
        """
        self.assertWarns(
            DeprecationWarning,
            "Use the sharing system to provide public pages, not IPublicPage",
            __file__,
            OfferingsFragmentTestCase.test_offerings, self)



class PublicFrontPageTests(TestCase, CommandStubMixin):
    """
    Tests for Mantissa's top-level web resource.
    """
    def setUp(self):
        """
        Set up a store with an installed offering.
        """
        self.siteStore = Store(dbdir=self.mktemp())
        Mantissa().installSite(self.siteStore, u"localhost", u"", False)
        off = Offering(
            name=u'test_offering',
            description=u'Offering for creating a sample app store',
            siteRequirements=[],
            appPowerups=[TestAppPowerup],
            installablePowerups=[],
            loginInterfaces=[],
            themes=[],
            )
        self.installedOffering = installOffering(self.siteStore, off, None)
        self.app = self.installedOffering.application
        self.substore = self.app.open()
        sharedItem = getEveryoneRole(self.substore).getShare(
            getDefaultShareID(self.substore))
        self.frontPage = self.siteStore.findUnique(FrontPage)
        self.webViewer = IWebViewer(self.siteStore)


    def test_offeringChild(self):
        """
        Installing an offering makes its shared items accessible under a child
        of L{_PublicFrontPage} with the offering's name.
        """
        frontPage = FrontPage(store=self.siteStore)
        resource = _PublicFrontPage(frontPage, self.webViewer)
        request = FakeRequest()
        result, segments = resource.locateChild(request, ('test_offering',))
        self.assertIdentical(result.userStore, self.substore)
        self.assertTrue(IWebViewer.providedBy(result.webViewer))


    def test_nonExistentChild(self):
        """
        L{_PublicFrontPage.locateChild} returns L{rend.NotFound} for a child
        segment which does not exist.
        """
        store = Store()
        frontPage = FrontPage(store=store)
        resource = _PublicFrontPage(frontPage, IWebViewer(self.siteStore))

        request = FakeRequest()
        ctx = context.WebContext()
        ctx.remember(request, inevow.IRequest)

        result = resource.locateChild(ctx, ('foo',))
        self.assertIdentical(result, rend.NotFound)


    def test_rootChild(self):
        """
        When no default offering has been selected,
        L{PublicFrontPage.locateChild} returns an L{_OfferingsFragment} wrapped by
        the L{IWebViewer}.
        """
        frontPage = FrontPage(store=self.siteStore)
        resource = _PublicFrontPage(frontPage, self.webViewer)
        request = FakeRequest()
        ctx = context.WebContext()
        ctx.remember(request, inevow.IRequest)
        result, segments = resource.locateChild(ctx, ('',))
        self.assertIsInstance(result, PublicPage)
        self.assertIsInstance(result.fragment, _OfferingsFragment)


    def test_rootChildWithDefaultApp(self):
        """
        The root resource provided by L{_PublicFrontPage} when a primary
        application has been selected is that application's L{SharingIndex}.
        """
        resource, segments = self.frontPage.produceResource(
            None, ('',), IWebViewer(self.siteStore))
        self.assertEqual(segments, ('',))
        self.frontPage.defaultApplication = self.app
        result, segments = resource.locateChild(None, ('',))
        self.assertIsInstance(result, PublicPage)
        self.assertIsInstance(result.fragment, TestAppElement)


    def getStore(self):
        return self.siteStore


    def test_switchFrontPage(self):
        """
        'axiomatic frontpage <offeringName>' switches the primary application
        (i.e., the one whose front page will be displayed on the site's root
        resource) to the one belonging to the named offering.
        """
        off2 = Offering(
            name=u'test_offering2',
            description=u'Offering for creating a sample app store',
            siteRequirements=[],
            appPowerups=[TestAppPowerup],
            installablePowerups=[],
            loginInterfaces=[],
            themes=[])
        installedOffering2 = installOffering(self.siteStore, off2, None)
        sfp = SetFrontPage()
        sfp.parent = self
        sfp.parseOptions(["test_offering"])
        resource, segments = self.frontPage.produceResource(
            None, ('',), self.webViewer)
        result, segs = resource.locateChild(None, [''])
        self.assertIdentical(result.fragment.original.store,
                             self.substore)
        self.assertEqual(segs, [])
        sfp.parseOptions(["test_offering2"])
        resource, moreSegs = self.frontPage.produceResource(None, ('',),
                                                            self.webViewer)
        result, segs = resource.locateChild(None, [''])
        self.assertEqual(segs, [])
        self.assertIdentical(result.fragment.original.store,
                             installedOffering2.application.open())

        self.assertRaises(UsageError, sfp.parseOptions, [])
        self.assertRaises(UsageError, sfp.parseOptions, ["nonexistent"])

class AuthenticatedNavigationTestMixin:
    """
    Mixin defining test methods for the authenticated navigation view.

    @ivar siteStore: The site Store with which the page returned by
        L{createPage} will be associated (probably by way of a user store and
        an item).
    """
    userinfo = (u'testuser', u'example.com')
    username = u'@'.join(userinfo)

    def createPage(self):
        """
        Create a subclass of L{_PublicPageMixin} to be used by tests.
        """
        raise NotImplementedError("%r did not implement createPage" % (self,))


    def rootURL(self, request):
        """
        Return the root URL for the website associated with the page returned
        by L{createPage}.
        """
        raise NotImplementedError("%r did not implement rootURL" % (self,))


    def test_headRenderer(self):
        """
        The I{head} renderer gets the head section for each installed theme by
        calling their C{head} method with the request being rendered and adds
        the result to the tag it is passed.
        """
        headCalls = []
        customHead = object()
        class CustomHeadTheme(object):
            priority = 0
            themeName = "custom"
            def head(self, request, site):
                headCalls.append((request, site))
                return customHead
            def getDocFactory(self, name, default):
                return default

        tech = FakeOfferingTechnician()
        tech.installOffering(Offering(
                name=u'fake', description=u'', siteRequirements=[],
                appPowerups=[], installablePowerups=[], loginInterfaces=[],
                themes=[CustomHeadTheme()]))
        self.siteStore.inMemoryPowerUp(tech, IOfferingTechnician)

        # Flush the cache, which is global, else the above fake-out is
        # completely ignored.
        theThemeCache.emptyCache()

        page = self.createPage(self.username)
        tag = div()
        ctx = context.WebContext(tag=tag)
        request = FakeRequest()
        ctx.remember(request, inevow.IRequest)
        result = page.render_head(ctx, None)
        self.assertEqual(result.tagName, 'div')
        self.assertEqual(result.attributes, {})
        self.assertIn(customHead, result.children)
        self.assertEqual(
            headCalls, [(request, ISiteURLGenerator(self.siteStore))])


    def test_authenticatedAuthenticateLinks(self):
        """
        The I{authenticateLinks} renderer should remove the tag it is passed
        from the output if it is called on a L{_PublicPageMixin} being rendered
        for an authenticated user.
        """
        page = self.createPage(self.username)
        authenticateLinksPattern = div()
        ctx = context.WebContext(tag=authenticateLinksPattern)
        tag = page.render_authenticateLinks(ctx, None)
        self.assertEqual(tag, '')


    def test_authenticatedStartmenu(self):
        """
        The I{startmenu} renderer should add navigation elements to the tag it
        is passed if it is called on a L{_PublicPageMixin} being rendered for an
        authenticated user.
        """
        navigable = FakeNavigableElement(store=self.userStore)
        installOn(navigable, self.userStore)
        navigable.tabs = [Tab('foo', 123, 0, [Tab('bar', 432, 0)])]

        page = self.createPage(self.username)
        startMenuTag = div[
            h1(pattern='tab'),
            h2(pattern='subtabs')]

        ctx = context.WebContext(tag=startMenuTag)
        tag = page.render_startmenu(ctx, None)
        self.assertEqual(tag.tagName, 'div')
        self.assertEqual(tag.attributes, {})
        children = [child for child in tag.children if child.pattern is None]
        self.assertEqual(len(children), 0)
        # This structure seems overly complex.
        tabs = list(tag.slotData.pop('tabs'))
        self.assertEqual(len(tabs), 1)
        fooTab = tabs[0]
        self.assertEqual(fooTab.tagName, 'h1')
        self.assertEqual(fooTab.attributes, {})
        self.assertEqual(fooTab.children, [])
        self.assertEqual(fooTab.slotData['href'], self.privateApp.linkTo(123))
        self.assertEqual(fooTab.slotData['name'], 'foo')
        self.assertEqual(fooTab.slotData['kids'].tagName, 'h2')
        subtabs = list(fooTab.slotData['kids'].slotData['kids'])
        self.assertEqual(len(subtabs), 1)
        barTab = subtabs[0]
        self.assertEqual(barTab.tagName, 'h1')
        self.assertEqual(barTab.attributes, {})
        self.assertEqual(barTab.children, [])
        self.assertEqual(barTab.slotData['href'], self.privateApp.linkTo(432))
        self.assertEqual(barTab.slotData['name'], 'bar')
        self.assertEqual(barTab.slotData['kids'], '')


    def test_authenticatedSettingsLink(self):
        """
        The I{settingsLink} renderer should add the URL of the settings item to
        the tag it is passed if it is called on a L{_PublicPageMixin} being
        rendered for an authenticated user.
        """
        page = self.createPage(self.username)
        settingsLinkPattern = div()
        ctx = context.WebContext(tag=settingsLinkPattern)
        tag = page.render_settingsLink(ctx, None)
        self.assertEqual(tag.tagName, 'div')
        self.assertEqual(tag.attributes, {})
        self.assertEqual(
            tag.children,
            [self.privateApp.linkTo(
                    self.userStore.findUnique(PreferenceAggregator).storeID)])


    def test_authenticatedLogout(self):
        """
        The I{logout} renderer should return the tag it is passed if it is
        called on a L{_PublicPageMixin} being rendered for an authenticated
        user.
        """
        page = self.createPage(self.username)
        logoutPattern = div()
        ctx = context.WebContext(tag=logoutPattern)
        tag = page.render_logout(ctx, None)
        self.assertIdentical(logoutPattern, tag)


    def test_authenticatedApplicationNavigation(self):
        """
        The I{applicationNavigation} renderer should add primary navigation
        elements to the tag it is passed if it is called on a
        L{_PublicPageMixin} being rendered for an authenticated user.
        """
        navigable = FakeNavigableElement(store=self.userStore)
        installOn(navigable, self.userStore)
        navigable.tabs = [Tab('foo', 123, 0, [Tab('bar', 432, 0)])]
        request = FakeRequest()

        page = self.createPage(self.username)
        navigationPattern = div[
            span(id='app-tab', pattern='app-tab'),
            span(id='tab-contents', pattern='tab-contents')]
        ctx = context.WebContext(tag=navigationPattern)
        ctx.remember(request)
        tag = page.render_applicationNavigation(ctx, None)
        self.assertEqual(tag.tagName, 'div')
        self.assertEqual(tag.attributes, {})
        children = [child for child in tag.children if child.pattern is None]
        self.assertEqual(children, [])
        self.assertEqual(len(tag.slotData['tabs']), 1)
        fooTab = tag.slotData['tabs'][0]
        self.assertEqual(fooTab.attributes, {'id': 'app-tab'})
        self.assertEqual(fooTab.slotData['name'], 'foo')
        fooContent = fooTab.slotData['tab-contents']
        self.assertEqual(fooContent.attributes, {'id': 'tab-contents'})
        self.assertEqual(
            fooContent.slotData['href'], self.privateApp.linkTo(123))


    def test_title(self):
        """
        The I{title} renderer should add the wrapped fragment's title
        attribute, if any, or the default "Divmod".
        """
        page = self.createPage(self.username)
        titleTag = title()
        tag = page.render_title(context.WebContext(tag=titleTag), None)
        self.assertIdentical(tag, titleTag)
        flattened = flatten(tag)
        self.assertSubstring(flatten(getattr(page.fragment, 'title', 'Divmod')),
                             flattened)


    def test_rootURL(self):
        """
        The I{base} renderer should add the website's root URL to the tag it is
        passed.
        """
        page = self.createPage(self.username)
        baseTag = div()
        request = FakeRequest(headers={'host': 'example.com'})
        ctx = context.WebContext(tag=baseTag)
        ctx.remember(request, inevow.IRequest)
        tag = page.render_rootURL(ctx, None)
        self.assertIdentical(tag, baseTag)
        self.assertEqual(tag.attributes, {})
        self.assertEqual(tag.children, [self.rootURL(request)])


    def test_noUsername(self):
        """
        The I{username} renderer should remove its node from the output when
        presented with a None username.
        """
        page = self.createPage(None)
        result = page.render_username(None, None)
        self.assertEqual(result, "")


    def test_noUrchin(self):
        """
        When there's no Urchin API key installed, the I{urchin} renderer should
        remove its node from the output.
        """
        page = self.createPage(None)
        result = page.render_urchin(None, None)
        self.assertEqual(result, "")


    def test_urchin(self):
        """
        When an Urchin API key is present, the code for enabling Google
        Analytics tracking should be inserted into the shell template.
        """
        keyString = u"UA-99018-11"
        APIKey.setKeyForAPI(self.siteStore, APIKey.URCHIN, keyString)
        page = self.createPage(None)
        t = div()
        result = page.render_urchin(context.WebContext(tag=t), None)
        self.assertEqual(result.slotData['urchin-key'], keyString)


    def usernameRenderingTest(self, username, hostHeader, expectedUsername):
        """
        Verify that the username will be rendered appropriately given the host
        of the HTTP request.

        @param username: the user's full login identifier.
        @param hostHeader: the value of the 'host' header.
        @param expectedUsername: the expected value of the rendered username.
        """
        page = self.createPage(username)
        userTag = span()
        req = FakeRequest(headers={'host': hostHeader})
        ctx = context.WebContext(tag=userTag)
        ctx.remember(req, inevow.IRequest)
        tag = page.render_username(ctx, None)
        self.assertEqual(tag.tagName, 'span')
        self.assertEqual(tag.children, [expectedUsername])


    def test_localUsername(self):
        """
        The I{username} renderer should render just the username when the
        username domain is the same as the HTTP request's domain. otherwise it
        should render the full username complete with domain.
        """
        domainUser = self.username.split('@')[0]
        return self.usernameRenderingTest(
            self.username, 'example.com', domainUser)


    def test_remoteUsername(self):
        """
        The I{username} renderer should render username with the domain when
        the username domain is different than the HTTP request's domain.
        """
        return self.usernameRenderingTest(
            self.username, 'not-example.com', self.username)


    def test_usernameWithHostPort(self):
        """
        The I{username} renderer should respect ports in the host headers.
        """
        domainUser = self.username.split('@')[0]
        return self.usernameRenderingTest(
            self.username, 'example.com:8080', domainUser)


    def test_prefixedDomainUsername(self):
        """
        The I{username} renderer should render just the username in the case
        where you are viewing a subdomain as well; if bob is viewing
        'jethro.divmod.com' or 'www.divmod.com', he should still see the
        username 'bob'.
        """
        domainUser = self.username.split('@')[0]
        return self.usernameRenderingTest(
            self.username, 'www.example.com', domainUser)



class _PublicAthenaLivePageTestMixin(AuthenticatedNavigationTestMixin):
    """
    Mixin which defines test methods which exercise functionality provided by
    the various L{xmantissa.publicweb._PublicPageMixin} subclasses, like
    L{PublicAthenaLivePage} and L{PublicNavAthenaLivePage}.
    """
    signupURL = u'sign/up'
    signupPrompt = u'sign up now'

    def setUp(self):
        self.siteStore = Store(filesdir=self.mktemp())

        def siteStoreTxn():
            Mantissa().installSite(
                self.siteStore, self.userinfo[1], u"", False)
            ticketed = signup.FreeTicketSignup(
                store=self.siteStore, prefixURL=self.signupURL,
                prompt=self.signupPrompt)
            signup._SignupTracker(store=self.siteStore, signupItem=ticketed)

            return  Create().addAccount(
                self.siteStore, self.userinfo[0],
                self.userinfo[1], u'password').avatars.open()

        self.userStore = self.siteStore.transact(siteStoreTxn)

        def userStoreTxn():
            self.privateApp = PrivateApplication(store=self.userStore)
            installOn(self.privateApp, self.userStore)
        self.userStore.transact(userStoreTxn)


    def rootURL(self, request):
        """
        Return the root URL as reported by C{self.website}.
        """
        return ISiteURLGenerator(self.siteStore).rootURL(request)


    def test_unauthenticatedAuthenticateLinks(self):
        """
        The I{authenticateLinks} renderer should add login and signup links to
        the tag it is passed, if it is called on a L{_PublicPageMixin} being
        rendered for an unauthenticated user.
        """
        page = self.createPage(None)
        authenticateLinksPattern = div[span(pattern='signup-link')]
        ctx = context.WebContext(tag=authenticateLinksPattern)
        tag = page.render_authenticateLinks(ctx, None)
        self.assertEqual(tag.tagName, 'div')
        self.assertEqual(tag.attributes, {})
        children = [child for child in tag.children if child.pattern is None]
        self.assertEqual(len(children), 1)
        self.assertEqual(
            children[0].slotData,
            {'prompt': self.signupPrompt, 'url': '/' + self.signupURL})


    def test_unauthenticatedStartmenu(self):
        """
        The I{startmenu} renderer should remove the tag it is passed from the
        output if it is called on a L{_PublicPageMixin} being rendered for an
        unauthenticated user.
        """
        page = self.createPage(None)
        startMenuTag = div()
        ctx = context.WebContext(tag=startMenuTag)
        tag = page.render_startmenu(ctx, None)
        self.assertEqual(tag, '')


    def test_unauthenticatedSettingsLink(self):
        """
        The I{settingsLink} renderer should remove the tag it is passed from
        the output if it is called on a L{_PublicPageMixin} being rendered for
        an unauthenticated user.
        """
        page = self.createPage(None)
        settingsLinkPattern = div()
        ctx = context.WebContext(tag=settingsLinkPattern)
        tag = page.render_settingsLink(ctx, None)
        self.assertEqual(tag, '')


    def test_unauthenticatedLogout(self):
        """
        The I{logout} renderer should remove the tag it is passed from the
        output if it is called on a L{_PublicPageMixin} being rendered for an
        authenticated user.
        """
        page = self.createPage(None)
        logoutPattern = div()
        ctx = context.WebContext(tag=logoutPattern)
        tag = page.render_logout(ctx, None)
        self.assertEqual(tag, '')


    def test_unauthenticatedApplicationNavigation(self):
        """
        The I{applicationNavigation} renderer should remove the tag it is
        passed from the output if it is called on a L{_PublicPageMixin} being
        rendered for an unauthenticated user.
        """
        page = self.createPage(None)
        navigationPattern = div()
        ctx = context.WebContext(tag=navigationPattern)
        tag = page.render_applicationNavigation(ctx, None)
        self.assertEqual(tag, '')



class TestFragment(rend.Fragment):
    title = u'a test fragment'



class PublicAthenaLivePageTestCase(_PublicAthenaLivePageTestMixin, TestCase):
    """
    Tests for L{PublicAthenaLivePage}.
    """
    def createPage(self, forUser):
        return PublicAthenaLivePage(
            self.siteStore, TestFragment(), forUser=forUser)



class PublicNavAthenaLivePageTestCase(_PublicAthenaLivePageTestMixin, TestCase):
    """
    Tests for L{PublicNavAthenaLivePage}.
    """
    suppress = [SUPPRESS(category=DeprecationWarning)]

    def createPage(self, forUser):
        return PublicNavAthenaLivePage(
            self.siteStore, TestFragment(), forUser=forUser)



class GetLoaderTests(TestCase):
    """
    Tests for L{xmantissa.publicweb.getLoader}.
    """
    def test_deprecated(self):
        """
        Calling L{getLoader} emits a deprecation warning.
        """
        self.assertWarns(
            DeprecationWarning,
            "xmantissa.publicweb.getLoader is deprecated, use "
            "PrivateApplication.getDocFactory or SiteTemplateResolver."
            "getDocFactory.",
            __file__,
            lambda: getLoader("shell"))



class CustomizedPublicPageTests(TestCase):
    """
    Tests for L{CustomizedPublicPage}.

    Let's say you've got a normal Mantissa database.  There's an
    L{AnonymousSite} in the site store, powering it up for L{IResource}.
    There's a user, that has a user store, which has a L{WebSite} as their
    L{IResource} avatar, plus a L{PrivateApplication} and a
    L{CustomizedPublicPage} as L{ISiteRootPlugin}s.

    L{CustomizedPublicPage}'s purpose is to make sure that when the user views
    the public site, their L{IWebViewer} is propagated to children of
    the global L{AnonymousSite}.
    """


    def setUp(self):
        """
        Create a store as described in the test case docstring.
        """
        site = Store(self.mktemp())
        Mantissa().installSite(site, u"example.com", u"", False)
        Mantissa().installAdmin(site, u'admin', u'example.com', u'asdf')
        anonsite = site.findUnique(AnonymousSite)
        user = site.findUnique(LoginSystem).accountByAddress(
            u"admin",u"example.com").avatars.open()
        self.website = user.findUnique(WebSite)
        self.privapp = user.findUnique(PrivateApplication)
        self.site = site


    def test_propagateNavigationToSlashUsers(self):
        """
        When the 'users' child is requested through a CustomizedPublicPage,
        L{AnonymousSite.rootChild_users} method should be invoked to produce a
        L{UserIndexPage} for the given user's L{PrivateApplication}.
        """
        wrapper, resultSegs = self.website.locateChild(
            FakeRequest(headers={"host": "example.com"}),
            ('users',))
        self.assertIsInstance(wrapper, _CustomizingResource)
        self.assertIsInstance(wrapper.currentResource, UserIndexPage)
        self.assertIsInstance(wrapper.currentResource.webViewer,
                              _AuthenticatedWebViewer)
        self.assertIdentical(wrapper.currentResource.webViewer._privateApplication,
                             self.privapp)


    def test_propagateNavigationToPlugins(self):
        """
        The site store has an L{ISiteRootPlugin} which provides some other
        application-defined resource - we'll call that 'AppSitePlugin'.

        L{CustomizedPublicPage}'s whole purpose is to make sure that when
        'AppSitePlugin' wants to return a resource, that resource is
        appropriately decorated so that its shell template will appear
        appropriate to the logged-in user.  In order to do that,
        'AppSitePlugin' must receive as its C{webViewer} argument the same
        one that L{CustomizedPublicPage} does.
        """
        # Stock configuration now set up, let's introduce a site plugin...

        result = object()
        calledWith = []
        class AppSitePlugin(object):
            implements(ISiteRootPlugin)
            def produceResource(self, request, segments, webViewer):
                calledWith.append([request, segments, webViewer])
                return result, ()

        self.site.inMemoryPowerUp(AppSitePlugin(), ISiteRootPlugin)
        req = FakeRequest(headers={"host": "example.com"})
        wrapper, resultSegs = self.website.locateChild(req, ("foo", "bar"))

        [(inreq, segs, webViewer)] = calledWith
        self.assertIdentical(inreq, req)
        self.assertEqual(segs, ('foo', 'bar'))
        self.assertIsInstance(webViewer, _AuthenticatedWebViewer)
        self.assertEqual(webViewer._privateApplication, self.privapp)

        self.assertIsInstance(wrapper, _CustomizingResource)
        self.assertIdentical(wrapper.currentResource, result)
        self.assertIdentical(resultSegs, ())



class AnonymousSiteTests(SiteTestsMixin, TestCase):
    """
    Tests for L{AnonymousSite}.
    """
    def setUp(self):
        """
        Set up a store with a valid offering to test against.
        """
        SiteTestsMixin.setUp(self)
        self.store = self.siteStore
        self.site = ISiteURLGenerator(self.store)
        self.resource = IResource(self.store)


    def test_powersUpWebViewer(self):
        """
        L{AnonymousSite} provides an indirected L{IWebViewer}
        powerup, and its indirected powerup should be the default provider of
        that interface.
        """
        webViewer = IWebViewer(self.store)
        self.assertIsInstance(webViewer, _AnonymousWebViewer)
        self.assertIdentical(webViewer._siteStore, self.store)


    def test_login(self):
        """
        L{AnonymousSite} has a I{login} child which returns a L{LoginPage}
        instance.
        """
        host = 'example.org'
        port = 1234
        netloc = '%s:%d' % (host, port)

        request = FakeRequest(
            headers={'host': netloc},
            uri='/login/foo',
            currentSegments=[],
            isSecure=False)

        self.site.hostname = host.decode('ascii')
        SSLPort(store=self.store, portNumber=port, factory=self.site)

        resource, segments = self.resource.locateChild(request, ("login",))
        self.assertTrue(isinstance(resource, LoginPage))
        self.assertIdentical(resource.store, self.store)
        self.assertEqual(resource.segments, ())
        self.assertEqual(resource.arguments, {})
        self.assertEqual(segments, ())


    def test_resetPassword(self):
        """
        L{AnonymousSite} has a I{resetPassword} child which returns a
        L{PasswordResetResource} instance.
        """
        resource, segments = self.resource.locateChild(
            FakeRequest(headers={"host": "example.com"}),
            ("resetPassword",))
        self.assertTrue(isinstance(resource, PasswordResetResource))
        self.assertIdentical(resource.store, self.store)
        self.assertEqual(segments, ())


    def test_users(self):
        """
        L{AnonymousSite} has a I{users} child which returns a L{UserIndexPage}
        instance.
        """
        resource, segments = self.resource.locateChild(
            FakeRequest(headers={"host": "example.com"}), ("users",))
        self.assertTrue(isinstance(resource, UserIndexPage))
        self.assertIdentical(
            resource.loginSystem, self.store.findUnique(LoginSystem))
        self.assertEqual(segments, ())


    def test_notFound(self):
        """
        L{AnonymousSite.locateChild} returns L{NotFound} for requests it cannot
        find another response for.
        """
        result = self.resource.locateChild(
            FakeRequest(headers={"host": "example.com"}),
            ("foo", "bar"))
        self.assertIdentical(result, NotFound)



from nevow.livetrial import testcase
from nevow.athena import expose
from xmantissa.signup import ValidatingSignupForm



class FakeUserInfoSignup:

    def createUser(self, firstName, lastName, username, domain,
                   password, emailAddress):
        assert False, "Form shouldn't be submitted"

    def usernameAvailable(self, username, domain):
        if username == 'bad':
            return [False, u'bad username']
        return [True, u"no reason"]



class TestUserInfoSignup(testcase.TestCase):
    jsClass = u'Mantissa.Test.UserInfoSignup'

    def getWidgetDocument(self):
        uis = FakeUserInfoSignup()
        vsf = ValidatingSignupForm(uis)
        vsf.setFragmentParent(self)
        return vsf



class TestSignupLocalpartValidation(TestUserInfoSignup):
    jsClass = u'Mantissa.Test.SignupLocalpartValidation'



class TestSignupValidationInformation(testcase.TestCase):
    jsClass = u'Mantissa.Test.SignupValidationInformation'

    def makeWidget(self):
        uis = FakeUserInfoSignup()
        vsf = ValidatingSignupForm(uis)
        vsf.setFragmentParent(self)
        return vsf
    expose(makeWidget)



# Copyright 2009 Divmod, Inc.  See LICENSE file for details

"""
Tests for L{xmantissa.terminal}.
"""

from zope.interface import implements
from zope.interface.verify import verifyObject

from twisted.internet.protocol import ProcessProtocol
from twisted.trial.unittest import TestCase
from twisted.cred.credentials import UsernamePassword
from twisted.conch.interfaces import IConchUser, ISession
from twisted.conch.ssh.keys import Key
from twisted.conch.ssh.session import SSHSession
from twisted.conch.insults.helper import TerminalBuffer
from twisted.conch.insults.insults import ServerProtocol
from twisted.conch.manhole import ColoredManhole

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import text, inmemory
from axiom.dependency import installOn
from axiom.userbase import LoginSystem, LoginMethod

from xmantissa.ixmantissa import IProtocolFactoryFactory, ITerminalServerFactory
from xmantissa.ixmantissa import IViewer
from xmantissa.sharing import getSelfRole
from xmantissa.terminal import SecureShellConfiguration, TerminalManhole
from xmantissa.terminal import ShellAccount, ShellServer, _ReturnToMenuWrapper
from xmantissa.terminal import _AuthenticatedShellViewer


class SecureShellConfigurationTests(TestCase):
    """
    Tests for L{xmantissa.shell.SecureShellConfiguration} which defines how to
    create an SSH server.
    """
    _hostKey = (
        "-----BEGIN RSA PRIVATE KEY-----\n"
        "MIIByAIBAAJhAM/dftm59mJJ1JVy0bsq8J7fp4WUecgaJukRyf637d76ywxRYGdw\n"
        "47hkBiJaDYgaE9HMlh2eSow3b2YCyom4FLlh/7Buq58A9IofR7ZiNVYv0ZDppbDg\n"
        "FN+Gl2ZFLFB3dwIBIwJgC+DFa4b4It+lv2Wllaquqf4m1G7iYzSxxCzm+JzLw5lN\n"
        "bmsM0rX+Yk7bx3LcM6m34vyvhY6p/kQyjHo7/CkpaSQg4bnpOcqEq3oMf8E0c0lp\n"
        "TQ1TdtfnKKrZZPTaVr7rAjEA7O19/tSLK6by1BpE1cb6W07GK1WcafYLxQLT64o+\n"
        "GKxbrlsossc8gWJ8GDRjE2S5AjEA4JkYfYkgfucH941r9yDFrhr6FuOdwbLXDESZ\n"
        "DyLhW/7DHiVIXlaLFnY+51PcTwWvAjBzFESDFsdBFpMz0j6w+j8WaBccXMhQuVYs\n"
        "fbdjxs20NnWsdWuKCQAhljxGRVSxpfMCMBmrGL3jyTMTFtp2j/78bl0KZbl5GVf3\n"
        "LoUPJ29xs1r4i1PnAPTWsM9d+I93TGDNcwIxAMRz4KO02tiLXG2igwDw/WWszrkr\n"
        "r4ggaFDlt4QqoNz0l4tayqzbDV1XceLgP4cXcQ==\n"
        "-----END RSA PRIVATE KEY-----\n")

    def setUp(self):
        """
        Create an in-memory L{Store} with a L{SecureShellConfiguration} in it.
        """
        self.store = Store()
        self.shell = SecureShellConfiguration(
            store=self.store, hostKey=self._hostKey)
        installOn(self.shell, self.store)


    def test_interfaces(self):
        """
        L{SecureShellConfiguration} implements L{IProtocolFactoryFactory}.
        """
        self.assertTrue(verifyObject(IProtocolFactoryFactory, self.shell))


    def test_powerup(self):
        """
        L{installOn} powers up the target for L{IProtocolFactoryFactory} with
        L{SecureShellConfiguration}.
        """
        self.assertIn(
            self.shell, list(self.store.powerupsFor(IProtocolFactoryFactory)))


    def test_repr(self):
        """
        The result of C{repr} on a L{SecureShellConfiguration} instance
        includes only a fingerprint of the private key, not the entire value.
        """
        self.assertEqual(
            repr(self.shell),
            "SecureShellConfiguration(storeID=%d, " % (self.shell.storeID,) +
            "hostKeyFingerprint='68cc7060bb6394060672467e7c4d8f3b')")


    def assertHostKey(self, shell, factory):
        """
        Assert that the public and private keys provided by C{factory}
        match those specified by C{shell} and that they are L{Key}
        instances.
        """
        privateKey = Key.fromString(shell.hostKey)
        self.assertEqual(
            factory.publicKeys, {'ssh-rsa': privateKey.public()})
        self.assertEqual(factory.privateKeys, {'ssh-rsa': privateKey})


    def test_getFactory(self):
        """
        L{SecureShellConfiguration.getFactory} returns an L{SSHFactory} with
        keys from L{SecureShellConfiguration.hostKey}.
        """
        factory = self.shell.getFactory()
        self.assertHostKey(self.shell, factory)


    def test_keyGeneration(self):
        """
        L{SecureShellConfiguration} generates its own key pair if one is not
        supplied to C{__init__}.
        """
        store = Store()
        shell = SecureShellConfiguration(store=store)
        installOn(shell, store)
        factory = shell.getFactory()
        self.assertHostKey(shell, factory)


    def test_portal(self):
        """
        The factory returned by L{SecureShellConfiguration.getFactory} has a
        C{portal} attribute which allows logins authenticated in the usual
        L{axiom.userbase} manner.
        """
        localpart = u'foo bar'
        domain = u'example.com'
        password = u'baz quux'

        loginSystem = self.store.findUnique(LoginSystem)
        account = loginSystem.addAccount(
            localpart, domain, password, internal=True)
        subStore = account.avatars.open()
        avatar = object()
        subStore.inMemoryPowerUp(avatar, IConchUser)
        factory = self.shell.getFactory()
        login = factory.portal.login(
            UsernamePassword(
                '%s@%s' % (localpart.encode('ascii'), domain.encode('ascii')),
                password),
            None, IConchUser)
        def cbLoggedIn(result):
            self.assertIdentical(IConchUser, result[0])
            self.assertIdentical(avatar, result[1])
        login.addCallback(cbLoggedIn)
        return login



class AuthenticatedShellViewerTests(TestCase):
    """
    Tests for L{_AuthenticatedShellViewer}, an L{IViewer} implementation for
    use with L{ITerminalServerFactory.buildTerminalProtocol}.
    """
    def test_interface(self):
        """
        L{_AuthenticatedShellViewer} instances provide L{IViewer}.
        """
        self.assertTrue(verifyObject(IViewer, _AuthenticatedShellViewer([])))


    def test_roleIn(self):
        """
        L{_AuthenticatedShellViewer.roleIn} returns a L{Role} for one of the
        account names passed to L{_AuthenticatedShellViewer.__init__}.
        """
        store = Store()
        viewer = _AuthenticatedShellViewer([(u"alice", u"example.com")])
        role = viewer.roleIn(store)
        self.assertEquals(role.externalID, u"alice@example.com")
        self.assertIdentical(role.store, store)



class ShellAccountTests(TestCase):
    """
    Tests for L{ShellAccount} which provide a basic L{IConchUser} avatar.
    """
    def setUp(self):
        """
        Create an in-memory L{Store} with a L{ShellAccount} in it.
        """
        self.store = Store()
        self.account = ShellAccount(store=self.store)
        installOn(self.account, self.store)


    def test_interfaces(self):
        """
        L{ShellAccount} powers up the item on which it is installed for
        L{IConchUser} and the L{IConchUser} powerup is adaptable to
        L{ISession}.
        """
        avatar = IConchUser(self.store)
        self.assertTrue(verifyObject(IConchUser, avatar))
        session = ISession(avatar)
        self.assertTrue(verifyObject(ISession, session))


    def test_lookupSessionChannel(self):
        """
        L{ShellAccount.lookupChannel} returns an L{SSHSession} instance.  (This
        is because L{SSHSession} implements handlers for the standard SSH
        requests issued to set up a shell.)
        """
        avatar = IConchUser(self.store)
        channel = avatar.lookupChannel('session', 65536, 16384, '')
        self.assertTrue(isinstance(channel, SSHSession))


    def test_openShell(self):
        """
        The L{ISession} adapter of the L{IConchUser} powerup implements
        C{openShell} so as to associate the given L{IProcessProtocol} with a
        transport.
        """
        proto = ProcessProtocol()
        session = ISession(IConchUser(self.store))

        # XXX See Twisted ticket #3864
        proto.session = session
        proto.write = lambda bytes: None

        # XXX See #2895.
        session.getPty(None, (123, 456, 789, 1000), None)
        session.openShell(proto)
        self.assertNotIdentical(proto.transport, None)



class FakeTerminal(TerminalBuffer):
    """
    A fake implementation of L{ITerminalTransport} used by the
    L{_ReturnToMenuWrapper} tests.
    """
    disconnected = False

    def loseConnection(self):
        self.disconnected = True



class ReturnToMenuTests(TestCase):
    """
    Tests for L{_ReturnToMenuWrapper} which wraps an L{ITerminalTransport} for
    an L{ITerminalProtocol} and switches to another L{ITerminalProtocol} when
    C{loseConnection} is called on it instead of disconnecting.
    """
    def test_write(self):
        """
        L{_ReturnToMenuWrapper.write} passes through to the wrapped terminal.
        """
        terminal = FakeTerminal()
        terminal.makeConnection(None)
        wrapper = _ReturnToMenuWrapper(None, terminal)
        wrapper.write('some bytes')
        wrapper.write('some more')
        self.assertIn('some bytessome more', str(terminal))


    def test_loseConnection(self):
        """
        L{_ReturnToMenuWrapper.loseConnection} does not disconnect the
        terminal; instead it calls the C{reactivate} method of its C{shell}
        attribute.
        """
        class FakeShell(object):
            activated = False

            def reactivate(self):
                self.activated = True

        shell = FakeShell()
        terminal = FakeTerminal()
        wrapper = _ReturnToMenuWrapper(shell, terminal)
        wrapper.loseConnection()
        self.assertFalse(terminal.disconnected)
        self.assertTrue(shell.activated)



class MockTerminalProtocol(object):
    """
    Implementation of L{ITerminalProtocol} used in test L{ShellServer}'s
    interactions with the interface.

    @ivar terminal: The L{ITerminalTransport} passed to C{makeConnection}.

    @ivar keystrokes: A C{list} of two-tuples giving each keystroke which this
        protocol has received.

    @ivar disconnected: A C{bool} indicating whether C{connectionLost} has been
        called yet.
    """
    def __init__(self):
        self.keystrokes = []
        self.terminal = None
        self.disconnected = False


    def makeConnection(self, terminal):
        self.terminal = terminal


    def connectionLost(self, reason):
        self.disconnected = True


    def keystrokeReceived(self, keyID, modifier):
        self.keystrokes.append((keyID, modifier))



class MockTerminalServerFactory(object):
    """
    Implementation of L{ITerminalServerFactory} used in test L{ShellServer}'s
    interactions with the interface.

    @ivar terminalProtocolInstance: The L{MockTerminalServer} created and
        returned by C{buildTerminalProtocol}, or C{None} if that method has not
        been called.
    """
    implements(ITerminalServerFactory)

    name = "mock"
    shellViewer = None
    terminalProtocolInstance = None

    def buildTerminalProtocol(self, shellViewer):
        self.shellViewer = shellViewer
        self.terminalProtocolInstance = MockTerminalProtocol()
        return self.terminalProtocolInstance

# Sanity check - this isn't a comprehensive (or even close) verification of
# MockTerminalServerFactory, but it at least points out obvious mistakes.
verifyObject(ITerminalServerFactory, MockTerminalServerFactory())



class MockTerminalServerFactoryItem(Item):
    """
    An L{Item} implementation of L{ITerminalServerFactory} used by tests.
    """
    powerupInterfaces = (ITerminalServerFactory,)
    implements(*powerupInterfaces)

    name = text()
    shellViewer = inmemory(
        doc="""
        The L{IViewer} passed to L{buildTerminalProtocol}.
        """)
    terminalProtocolInstance = inmemory(
        doc="""
        The L{MockTerminalServer} created and returned by
        C{buildTerminalProtocol}, or C{None} if that method has not been
        called.
        """)

    def activate(self):
        self.shellViewer = None
        self.terminalProtocolInstance = None


    def buildTerminalProtocol(self, shellViewer):
        self.shellViewer = shellViewer
        self.terminalProtocolInstance = MockTerminalProtocol()
        return self.terminalProtocolInstance

# Sanity check - see above call to verifyObject.
verifyObject(ITerminalServerFactory, MockTerminalServerFactoryItem())



class ShellServerTests(TestCase):
    """
    Tests for L{ShellServer} which is the top-level L{ITerminalProtocol},
    interacting initially and directly with terminals by presenting a menu of
    possible activities and delegating to other L{ITerminalProtocol}s which
    appropriate.
    """
    def test_switchTo(self):
        """
        L{ShellServer.switchTo} takes a L{ITerminalServerFactory} and uses it
        to create a new L{ITerminalProtocol} which it connects to a
        L{_ReturnToMenuWrapper}.  L{buildTerminalProtocol} is passed an
        L{IViewer}.
        """
        terminal = FakeTerminal()
        store = Store()
        # Put a login method into the store so it can have a role.  See #2665.
        LoginMethod(
            store=store, internal=True, protocol=u'*', verified=True,
            localpart=u'alice', domain=u'example.com',
            # Not really an account, but simpler...
            account=store)
        server = ShellServer(store)
        server.makeConnection(terminal)
        factory = MockTerminalServerFactory()
        server.switchTo(factory)
        self.assertIdentical(factory.shellViewer.roleIn(store), getSelfRole(store))
        self.assertTrue(isinstance(server._protocol, MockTerminalProtocol))
        self.assertTrue(isinstance(server._protocol.terminal, _ReturnToMenuWrapper))
        self.assertIdentical(server._protocol.terminal._shell, server)
        self.assertIdentical(server._protocol.terminal._terminal, terminal)


    def test_appButtons(self):
        """
        L{ShellServer._appButtons} returns an iterator the elements of which
        are L{Button} instances, one for each L{ITerminalServerFactory}
        powerup.  When one of these buttons is activated, L{ShellServer} is
        switched to the corresponding L{ITerminalServerFactory}'s protocol.
        """
        store = Store()
        terminal = FakeTerminal()
        server = ShellServer(store)
        server.makeConnection(terminal)

        firstFactory = MockTerminalServerFactoryItem(
            store=store, name=u"first - \N{ROMAN NUMERAL ONE}")
        installOn(firstFactory, store)
        secondFactory = MockTerminalServerFactoryItem(
            store=store, name=u"second - \N{ROMAN NUMERAL TWO}")
        installOn(secondFactory, store)

        buttons = list(server._appButtons())
        self.assertEqual(len(buttons), 2)

        # For now, we'll say the order isn't significant.
        buttons.sort(key=lambda b: b.label)

        self.assertEqual(
            buttons[0].label, firstFactory.name.encode('utf-8'))
        buttons[0].onPress()

        server.keystrokeReceived('x', None)
        self.assertEqual(
            firstFactory.terminalProtocolInstance.keystrokes, [('x', None)])

        self.assertEqual(
            buttons[1].label, secondFactory.name.encode('utf-8'))
        buttons[1].onPress()
        server.keystrokeReceived('y', None)
        self.assertEqual(
            secondFactory.terminalProtocolInstance.keystrokes, [('y', None)])


    def test_logoffButton(self):
        """
        L{ShellServer._logoffButton} returns a L{Button} which, when activated,
        disconnects the terminal.
        """
        terminal = FakeTerminal()
        server = ShellServer(Store())
        server.makeConnection(terminal)
        server._logoffButton().onPress()
        self.assertTrue(terminal.disconnected)


    def test_reactivate(self):
        """
        L{ShellServer.reactivate} disconnects the protocol previously switched
        to, drops the reference to it, and redraws the main menu.
        """
        terminal = FakeTerminal()
        server = ShellServer(Store())
        server.makeConnection(terminal)
        server.switchTo(MockTerminalServerFactory())
        server.reactivate()
        self.assertIdentical(server._protocol, None)


    def test_keystrokeReceivedWindow(self):
        """
        L{ShellServer.keystrokeReceived} delivers keystroke data to the main
        menu widget when no protocol has been switched to.
        """
        class FakeWidget(object):
            def __init__(self):
                self.keystrokes = []

            def keystrokeReceived(self, keyID, modifier):
                self.keystrokes.append((keyID, modifier))

        terminal = FakeTerminal()
        window = FakeWidget()
        server = ShellServer(Store())
        server._makeWindow = lambda: window
        server.makeConnection(terminal)
        server.keystrokeReceived(' ', ServerProtocol.ALT)
        self.assertEqual(window.keystrokes, [(' ', ServerProtocol.ALT)])


    def test_keystrokeReceivedProtocol(self):
        """
        L{ShellServer.keystrokeReceived} delivers keystroke data to the
        protocol built by the factory which has been switched to.
        """
        factory = MockTerminalServerFactory()
        terminal = FakeTerminal()
        server = ShellServer(Store())
        server.makeConnection(terminal)
        server.switchTo(factory)
        server.keystrokeReceived(' ', ServerProtocol.ALT)
        self.assertEqual(
            factory.terminalProtocolInstance.keystrokes,
            [(' ', ServerProtocol.ALT)])



class ManholeTests(TestCase):
    """
    Tests for L{TerminalManhole} which provides an L{ITerminalServerFactory}
    for a protocol which gives a user an in-process Python REPL.
    """
    def test_interface(self):
        """
        L{TerminalManhole} implements L{ITerminalServerFactory}.
        """
        self.assertTrue(verifyObject(ITerminalServerFactory, TerminalManhole()))


    def test_buildTerminalProtocol(self):
        """
        L{TerminalManhole.buildTerminalProtocol} returns a L{ColoredManhole}
        with a namespace including the store the L{TerminalManhole} is in.
        """
        store = Store()
        factory = TerminalManhole(store=store)
        viewer = object()
        protocol = factory.buildTerminalProtocol(viewer)
        self.assertTrue(isinstance(protocol, ColoredManhole))
        self.assertEqual(protocol.namespace, {'db': store, 'viewer': viewer})



"""
Helpful utilities for code which tests functionality related to
L{xmantissa.people}.
"""
from zope.interface import implements

from twisted.python.reflect import qual

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import inmemory, text

from xmantissa.ixmantissa import IPeopleFilter, IContactType, IWebTranslator
from xmantissa.people import Organizer

from epsilon.descriptor import requiredAttribute


class PeopleFilterTestMixin:
    """
    Mixin for testing L{IPeopleFilter} providers.  Requires the following
    attributes:

    @ivar peopleFilterClass: The L{IPeopleFilter} being tested.
    @type peopleFilterClass: L{IPeopleFilter} provider.

    @ivar peopleFilterName: The expected name of L{peopleFilterClass}.
    @type peopleFilterName: C{str}
    """
    peopleFilterClass = requiredAttribute('peopleFilterClass')
    peopleFilterName = requiredAttribute('peopleFilterName')


    def assertComparisonEquals(self, comparison):
        """
        Instantiate L{peopleFilterClass}, call
        L{IPeopleFilter.getPeopleQueryComparison} on it and assert that its
        result is equal to C{comparison}.

        @type comparison: L{axiom.iaxiom.IComparison}
        """
        peopleFilter = self.peopleFilterClass()
        actualComparison = peopleFilter.getPeopleQueryComparison(Store())
        # none of the Axiom query objects have meaningful equality
        # comparisons, but their string representations do.
        # this assertion should be addressed along with #2464
        self.assertEqual(str(actualComparison), str(comparison))


    def makeOrganizer(self):
        """
        Return an L{Organizer}.
        """
        return Organizer(store=Store())


    def test_implementsInterface(self):
        """
        Our people filter should provide L{IPeopleFilter}.
        """
        self.assertTrue(IPeopleFilter.providedBy(self.peopleFilterClass()))


    def test_organizerIncludesIt(self):
        """
        L{Organizer.getPeopleFilters} should include an instance of our
        L{IPeopleFilter}.
        """
        organizer = self.makeOrganizer()
        self.assertIn(
            self.peopleFilterClass,
            [filter.__class__ for filter in organizer.getPeopleFilters()])


    def test_filterName(self):
        """
        Our L{IPeopleFilter}'s I{filterName} should match L{peopleFilterName}.
        """
        self.assertEqual(
            self.peopleFilterClass().filterName, self.peopleFilterName)



class StubPerson(object):
    """
    Stub implementation of L{Person} used for testing.

    @ivar contactItems: A list of three-tuples of the arguments passed to
        createContactInfoItem.
    """
    name = u'person'

    def __init__(self, contactItems):
        self.contactItems = contactItems
        self.store = object()


    def createContactInfoItem(self, cls, attr, value):
        """
        Record the creation of a new contact item.
        """
        self.contactItems.append((cls, attr, value))


    def getContactInfoItems(self, itemType, valueColumn):
        """
        Return an empty list.
        """
        return []


    def getMugshot(self):
        """
        Return C{None} since there is no mugshot.
        """
        return None


    def getDisplayName(self):
        """
        Return a name of some sort.
        """
        return u"Alice"


    def getEmailAddress(self):
        """
        Return an email address.
        """
        return u"alice@example.com"



class StubTranslator(object):
    """
    Translate between a dummy row identifier and a dummy object.
    """
    implements(IWebTranslator)

    def __init__(self, rowIdentifier, item):
        self.fromWebID = {rowIdentifier: item}.__getitem__
        self.toWebID = {item: rowIdentifier}.__getitem__



class StubOrganizer(object):
    """
    Mimic some of the API presented by L{Organizer}.

    @ivar people: A C{dict} mapping C{unicode} strings giving person names to
    person objects.  These person objects will be returned from appropriate
    calls to L{personByName}.

    @ivar contactTypes: a C{list} of L{IContactType}s.

    @ivar groupedReadOnlyViews: The C{dict} to be returned from
    L{groupReadOnlyViews}

    @ivar editedPeople: A list of the arguments which have been passed to the
    C{editPerson} method.

    @ivar deletedPeople: A list of the arguments which have been passed to the
    C{deletePerson} method.

    @ivar contactEditorialParameters: A mapping of people to lists.  When
    passed a person, L{getContactEditorialParameters} will return the
    corresponding list.

    @ivar groupedReadOnlyViewPeople: A list of the arguments passed to
    L{groupReadOnlyViews}.

    @ivar peopleTags: The value to return from L{getPeopleTags}.
    @type peopleTags: C{list}

    @ivar peopleFilters: The sequence to return from L{getPeopleFilters}.
    @type peopleFilters: C{list}

    @ivar organizerPlugins: The sequence of return from L{getOrganizerPlugins}.
    @type organizerPlugins: C{list}
    """
    _webTranslator = StubTranslator(None, None)

    def __init__(self, store=None, contactTypes=None, deletedPeople=None,
            editedPeople=None, contactEditorialParameters=None,
            groupedReadOnlyViews=None, peopleTags=None, peopleFilters=None,
            organizerPlugins=None):
        self.store = store
        self.people = {}
        if contactTypes is None:
            contactTypes = []
        if deletedPeople is None:
            deletedPeople = []
        if editedPeople is None:
            editedPeople = []
        if contactEditorialParameters is None:
            contactEditorialParameters = []
        if groupedReadOnlyViews is None:
            groupedReadOnlyViews = {}
        if peopleTags is None:
            peopleTags = []
        if peopleFilters is None:
            peopleFilters = []
        if organizerPlugins is None:
            organizerPlugins = []
        self.contactTypes = contactTypes
        self.deletedPeople = deletedPeople
        self.editedPeople = editedPeople
        self.contactEditorialParameters = contactEditorialParameters
        self.groupedReadOnlyViews = groupedReadOnlyViews
        self.groupedReadOnlyViewPeople = []
        self.peopleTags = peopleTags
        self.peopleFilters = peopleFilters
        self.organizerPlugins = organizerPlugins


    def personByName(self, name):
        return self.people[name]


    def lastNameOrder(self):
        return None


    def deletePerson(self, person):
        self.deletedPeople.append(person)


    def editPerson(self, person, name, edits):
        self.editedPeople.append((person, name, edits))


    def toContactEditorialParameter(self, contactType, person):
        for (_contactType, param) in self.contactEditorialParameters[person]:
            if _contactType == contactType:
                return param


    def getContactEditorialParameters(self, person):
        return self.contactEditorialParameters[person]


    def getContactTypes(self):
        return self.contactTypes


    def getPeopleFilters(self):
        """
        Return L{peopleFilters}.
        """
        return self.peopleFilters


    def groupReadOnlyViews(self, person):
        """
        Return L{groupedReadOnlyViews}.
        """
        self.groupedReadOnlyViewPeople.append(person)
        return self.groupedReadOnlyViews


    def linkToPerson(self, person):
        return "/person/" + person.getDisplayName()


    def getPeopleTags(self):
        """
        Return L{peopleTags}.
        """
        return self.peopleTags


    def getOrganizerPlugins(self):
        """
        Return L{organizerPlugins}.
        """
        return self.organizerPlugins



class StubOrganizerPlugin(Item):
    """
    Organizer powerup which records which people are created and gives back
    canned responses to method calls.
    """
    name = text(
        doc="""
        @see IOrganizerPlugin.name
        """)

    createdPeople = inmemory(
        doc="""
        A list of all L{Person} items created since this item was last loaded
        from the database.
        """)

    contactTypes = inmemory(
        doc="""
        A list of L{IContactType} implementors to return from
        L{getContactTypes}.
        """)

    peopleFilters = inmemory(
        doc="""
        A list of L{IPeopleFilter} imlpementors to return from
        L{getPeopleFilters}.
        """)

    renamedPeople = inmemory(
        doc="""
        A list of two-tuples of C{unicode} with the first element giving the
        name of each L{Person} item whose name changed at the time of the
        change and the second element giving the value passed for the old name
        parameter.
        """)

    createdContactItems = inmemory(
        doc="""
        A list of contact items created since this item was last loaded from
        the database.
        """)

    editedContactItems = inmemory(
        doc="""
        A list of contact items edited since this item was last loaded from
        the database.
        """)

    personalization = inmemory(
        doc="""
        An objects to be returned by L{personalize}.
        """)

    personalizedPeople = inmemory(
        doc="""
        A list of people passed to L{personalize}.
        """)

    def activate(self):
        """
        Initialize in-memory state tracking attributes to default values.
        """
        self.createdPeople = []
        self.renamedPeople = []
        self.createdContactItems = []
        self.editedContactItems = []
        self.personalization = None
        self.personalizedPeople = []


    def personCreated(self, person):
        """
        Record the creation of a L{Person}.
        """
        self.createdPeople.append(person)


    def personNameChanged(self, person, oldName):
        """
        Record the change of a L{Person}'s name.
        """
        self.renamedPeople.append((person.name, oldName))


    def contactItemCreated(self, contactItem):
        """
        Record the creation of a contact item.
        """
        self.createdContactItems.append(contactItem)


    def contactItemEdited(self, contactItem):
        """
        Record the editing of a contact item.
        """
        self.editedContactItems.append(contactItem)


    def getContactTypes(self):
        """
        Return the contact types list this item was constructed with.
        """
        return self.contactTypes


    def getPeopleFilters(self):
        """
        Return L{peopleFilters}.
        """
        return self.peopleFilters


    def personalize(self, person):
        """
        Record a personalization attempt and return C{self.personalization}.
        """
        self.personalizedPeople.append(person)
        return self.personalization



class StubReadOnlyView(object):
    """
    Test double for the objects returned by L{IContactType.getReadOnlyView}.

    @ivar item: The contact item the view is for.
    @ivar type: The contact type the contact item comes from.
    """
    def __init__(self, contactItem, contactType):
        self.item = contactItem
        self.type = contactType




class StubContactType(object):
    """
    Behaviorless contact type implementation used for tests.

    @ivar parameters: A list of L{xmantissa.liveform.Parameter} instances which
        will become the return value of L{getParameters}.
    @ivar createdContacts: A list of tuples of the arguments passed to
        C{createContactItem}.
    @ivar editorialForm: The object which will be returned from
        L{getEditFormForPerson}.
    @ivar editedContacts: A list of the contact items passed to
        L{getEditFormForPerson}.
    @ivar contactItems: The list of objects which will be returned from
        L{getContactItems}.
    @ivar queriedPeople: A list of the person items passed to
        L{getContactItems}.
    @ivar editedContacts: A list of two-tuples of the arguments passed to
        L{editContactItem}.
    @ivar createContactItems: A boolean indicating whether C{createContactItem}
        will return an object pretending to be a new contact item (C{True}) or
        C{None} to indicate no contact item was created (C{False}).
    @ivar theDescriptiveIdentifier: The object to return from
        L{descriptiveIdentifier}.
    @ivar contactGroup: The object to return from L{getContactGroup}.
    """
    implements(IContactType)

    def __init__(self, parameters, editorialForm, contactItems,
            createContactItems=True, allowMultipleContactItems=True,
            theDescriptiveIdentifier=u'', contactGroup=None):
        self.parameters = parameters
        self.createdContacts = []
        self.editorialForm = editorialForm
        self.editedContacts = []
        self.contactItems = contactItems
        self.queriedPeople = []
        self.editedContacts = []
        self.createContactItems = createContactItems
        self.allowMultipleContactItems = allowMultipleContactItems
        self.theDescriptiveIdentifier = theDescriptiveIdentifier
        self.contactGroup = contactGroup


    def getParameters(self, ignore):
        """
        Return L{parameters}.
        """
        return self.parameters


    def uniqueIdentifier(self):
        """
        Return the L{qual} of this class.
        """
        return qual(self.__class__).decode('ascii')


    def descriptiveIdentifier(self):
        """
        Return L{theDescriptiveIdentifier}.
        """
        return self.theDescriptiveIdentifier


    def getEditFormForPerson(self, contact):
        """
        Return an object which is supposed to be a form for editing an existing
        instance of this contact type and record the contact object which was
        passed in.
        """
        self.editedContacts.append(contact)
        return self.editorialForm


    def createContactItem(self, person, **parameters):
        """
        Record an attempt to create a new contact item of this type for the
        given person.
        """
        contactItem = (person, parameters)
        self.createdContacts.append(contactItem)
        if self.createContactItems:
            return contactItem
        return None


    def getContactItems(self, person):
        """
        Return C{self.contactItems} and record the person item passed in.
        """
        self.queriedPeople.append(person)
        return self.contactItems


    def editContactItem(self, contact, **changes):
        """
        Record an attempt to edit the details of a contact item.
        """
        self.editedContacts.append((contact, changes))


    def getContactGroup(self, contactItem):
        """
        Return L{contactGroup}.
        """
        return self.contactGroup


    def getReadOnlyView(self, contact):
        """
        Return a stub view object for the given contact.

        @rtype: L{StubReadOnlyView}
        """
        return StubReadOnlyView(contact, self)




from twisted.trial import unittest

from axiom import store, userbase
from axiom.item import Item
from axiom.attributes import inmemory, integer
from axiom.plugins import mantissacmd

from xmantissa import signup, offering
from xmantissa.plugins import free_signup
from xmantissa.product import Product, Installation



class SignupCreationTestCase(unittest.TestCase):
    def setUp(self):
        self.store = store.Store()
        self.ls = userbase.LoginSystem(store=self.store)
        self.admin = self.ls.addAccount(u'admin', u'localhost', None,
                                        internal=True, verified=True)
        self.substore = self.admin.avatars.open()
        self.sc = signup.SignupConfiguration(store=self.substore)

    def _installTestOffering(self):
        io = offering.InstalledOffering(
            store=self.store,
            offeringName=u"mantissa",
            application=None)

    def createFreeSignup(self, itemClass, url=u'signup', prompt=u'Sign Up!'):
        """

        A utility method to ensure that the same arguments are always used to
        create signup mechanisms, since these are the arguments that are going
        to be coming from the admin form.

        """
        product = Product(store=self.store, types=[])
        return self.sc.createSignup(
            u'testuser@localhost',
            itemClass,
            {'prefixURL': url},
            product,
            u'Blank Email Template', prompt)

    def testCreateFreeSignups(self):
        self._installTestOffering()

        for signupMechanismPlugin in [free_signup.freeTicket,
                                      free_signup.userInfo]:
            self.createFreeSignup(signupMechanismPlugin.itemClass)



    def test_usernameAvailability(self):
        """
        Test that the usernames which ought to be available are and that those
        which aren't are not:

        Only syntactically valid localparts are allowed.  Localparts which are
        already assigned are not allowed.

        Only domains which are actually served by this mantissa instance are
        allowed.
        """

        signup = self.createFreeSignup(free_signup.userInfo.itemClass)
        # Allowed: unused localpart, same domain as the administrator created
        # by setUp.
        self.failUnless(signup.usernameAvailable(u'alice', u'localhost')[0])

        # Not allowed: unused localpart, unknown domain.
        self.failIf(signup.usernameAvailable(u'alice', u'example.com')[0])

        # Not allowed: used localpart, same domain as the administrator created
        # by setUp.
        self.failIf(signup.usernameAvailable(u'admin', u'localhost')[0])
        self.assertEquals(signup.usernameAvailable(u'fjones', u'localhost'),
                          [True, u'Username already taken'])

        signup.createUser(
            realName=u"Frank Jones",
            username=u'fjones',
            domain=u'localhost',
            password=u'asdf',
            emailAddress=u'fj@crappy.example.com')

        self.assertEquals(signup.usernameAvailable(u'fjones', u'localhost'),
                          [False, u'Username already taken'])
        ss = self.ls.accountByAddress(u"fjones", u"localhost").avatars.open()
        self.assertEquals(ss.query(Installation).count(), 1)


    def testUserInfoSignupValidation2(self):
        """
        Ensure that invalid characters aren't allowed in usernames, that
        usernames are parsable as the local part of an email address and that
        usernames shorter than two characters are invalid.
        """
        signup = self.createFreeSignup(free_signup.userInfo.itemClass)
        self.assertEquals(signup.usernameAvailable(u'foo bar', u'localhost'),
                          [False, u"Username contains invalid character: ' '"])
        self.assertEquals(signup.usernameAvailable(u'foo@bar', u'localhost'),
                          [False, u"Username contains invalid character: '@'"])
        # '~' is not expressly forbidden by the validator in usernameAvailable,
        # yet it is rejected by parseAddress (in xmantissa.smtp).
        self.assertEquals(signup.usernameAvailable(u'fo~o', u'127.0.0.1'),
                          [False, u"Username fails to parse"])
        self.assertEquals(signup.usernameAvailable(u'f', u'localhost'),
                          [False, u"Username too short"])


    def test_userInfoSignupUserInfo(self):
        """
        Check that C{createUser} creates a L{signup.UserInfo} item with its
        C{realName} attribute set.
        """
        freeSignup = self.createFreeSignup(free_signup.userInfo.itemClass)
        freeSignup.createUser(
            u'Frank Jones', u'fjones', u'divmod.com',
            u'asdf', u'fj@example.com')
        account = self.ls.accountByAddress(u'fjones', u'divmod.com')
        substore = account.avatars.open()
        userInfos = list(substore.query(signup.UserInfo))
        self.assertEqual(len(userInfos), 1)
        userInfo = userInfos[0]
        self.assertEqual(userInfo.realName, u'Frank Jones')


    def test_userInfoCreatedBeforeProductInstalled(self):
        """
        L{UserInfoSignup.createUser} should create a L{UserInfo} item B{before} it
        calls L{Product.installProductOn}.
        """
        class StubProduct(Item):
            """
            L{Product}-alike which records the existing L{UserInfo} items in
            the store when it is installed.
            """
            required_axiom_attribute_garbage = integer(
                doc="""
                mandatory Item attribute.
                """)

            userInfos = inmemory()

            def installProductOn(self, substore):
                """
                Find all the L{UserInfo} items in the given store and remember
                them.
                """
                self.userInfos = list(substore.query(signup.UserInfo))

        product = StubProduct(store=self.store)
        freeSignup = self.createFreeSignup(free_signup.userInfo.itemClass)
        freeSignup.product = product
        freeSignup.createUser(
            u'Frank Jones', u'fjones', u'example.com',
            u'password', u'fj@example.org')
        self.assertEqual(len(product.userInfos), 1)


    def test_userInfoLoginMethods(self):
        """
        Check that C{createUser} creates only two L{LoginMethod}s on the
        account.
        """
        username, domain = u'fjones', u'divmod.com'
        signup = self.createFreeSignup(free_signup.userInfo.itemClass)
        signup.createUser(u'Frank Jones', username, domain, u'asdf',
                          u'fj@example.com')
        account = self.ls.accountByAddress(username, domain)
        query = list(
            self.store.query(userbase.LoginMethod,
                             userbase.LoginMethod.account == account,
                             sort=userbase.LoginMethod.internal.ascending))
        self.assertEquals(len(query), 2)
        self.assertEquals(query[0].internal, False)
        self.assertEquals(query[0].verified, False)
        self.assertEquals(query[0].localpart, u'fj')
        self.assertEquals(query[0].domain, u'example.com')
        self.assertEquals(query[1].internal, True)
        self.assertEquals(query[1].verified, True)
        self.assertEquals(query[1].localpart, username)
        self.assertEquals(query[1].domain, domain)


    def test_freeSignupsList(self):
        """
        Test that if we produce 3 different publicly accessible signups, we get
        information about all of them back.
        """
        for i, signupMechanismPlugin in enumerate(
            [free_signup.freeTicket,
             free_signup.userInfo]):
            self.createFreeSignup(signupMechanismPlugin.itemClass,
                                  url=u'signup%d' % (i+1,),
                                  prompt=u"Sign Up %d" % (i+1,))
        x = list(signup._getPublicSignupInfo(self.store))
        x.sort()
        self.assertEquals(x, [(u'Sign Up 1', u'/signup1'),
                              (u'Sign Up 2', u'/signup2')])



class ValidatingSignupFormTests(unittest.TestCase):
    """
    Tests for L{ValidatingSignupForm}.
    """
    def test_getInitialArguments(self):
        """
        L{ValidatingSignupForm.getInitialArguments} should return a tuple
        consisting of a unicode string giving the domain name for which this
        form will allow signup.
        """
        domain = u"example.com"
        siteStore = store.Store(filesdir=self.mktemp())
        mantissacmd.Mantissa().installSite(siteStore, domain, u"", False)
        login = siteStore.findUnique(userbase.LoginSystem)
        login.addAccount(u"alice", domain, u"password", internal=True)
        userInfo = signup.UserInfoSignup(store=siteStore, prefixURL=u"opaque")
        form = signup.ValidatingSignupForm(userInfo)
        self.assertEqual(form.getInitialArguments(), (domain,))



from axiom.test.historic.stubloader import saveStub

from xmantissa.signup import FreeTicketSignup, Multifactor
from xmantissa.webadmin import AdministrativeBenefactor

def createDatabase(s):
    ab = AdministrativeBenefactor(store=s)
    mf = Multifactor(store=s)
    mf.add(ab)
    FreeTicketSignup(store=s,
                     prefixURL=u'/a/b',
                     booth=s,
                     benefactor=mf,
                     emailTemplate=u'TEMPLATE!',
                     prompt=u'OK?')

if __name__ == '__main__':
    saveStub(createDatabase, 10664)



from axiom.test.historic.stubloader import saveStub

from xmantissa.signup import FreeTicketSignup

def createDatabase(s):
    FreeTicketSignup(store=s,
                     prefixURL=u'/a/b',
                     booth=s,
                     benefactor=s,
                     emailTemplate=u'TEMPLATE!')

if __name__ == '__main__':
    saveStub(createDatabase, 5809)




import os, sys

from cStringIO import StringIO

from twisted.python.usage import UsageError
from twisted.trial.unittest import TestCase

from axiom.plugins import webcmd

from axiom.store import Store
from axiom.test.util import CommandStubMixin
from axiom.plugins.mantissacmd import Mantissa

from xmantissa.web import SiteConfiguration
from xmantissa.website import APIKey

def _captureStandardOutput(f, *a, **k):
    """
    Capture standard output produced during the invocation of a function, and
    return it.

    Since this is for testing command-line tools, SystemExit errors that
    indicate a successful return are caught.
    """
    io = StringIO()
    oldout = sys.stdout
    sys.stdout = io
    try:
        try:
            f(*a, **k)
        finally:
            sys.stdout = oldout
    except SystemExit, se:
        if se.args[0]:
            raise
    return io.getvalue()


class TestIdempotentListing(CommandStubMixin, TestCase):

    def setUp(self):
        self.store = Store()
        self.options = webcmd.WebConfiguration()
        self.options.parent = self


    def test_requiresBaseOffering(self):
        """
        L{WebConfiguration.postOptions} raises L{UsageError} if it is used on a
        store which does not have the Mantissa base offering installed.
        """
        self.assertRaises(UsageError, self.options.postOptions)


    def _list(self):
        wconf = webcmd.WebConfiguration()
        wconf.parent = self
        wout = _captureStandardOutput(wconf.parseOptions, ['--list'])
        return wout

    def testListDoesNothing(self):
        """
        Verify that 'axiomatic -d foo.axiom web --list' does not modify
        anything, by running it twice and verifying that the generated output
        is identical the first and second time.
        """
        self.assertEquals(self._list(),
                          self._list())


class ConfigurationTestCase(CommandStubMixin, TestCase):
    def setUp(self):
        self.store = Store(filesdir=self.mktemp())
        Mantissa().installSite(self.store, u"example.com", u"", False)


    def test_shortOptionParsing(self):
        """
        Test that the short form of all the supported command line options are
        parsed correctly.
        """
        opt = webcmd.WebConfiguration()
        opt.parent = self
        certFile = self.store.filesdir.child('name')
        opt.parseOptions(['-h', 'http.log', '-H', 'example.com'])
        self.assertEquals(opt['http-log'], 'http.log')
        self.assertEquals(opt['hostname'], 'example.com')


    def test_longOptionParsing(self):
        """
        Test that the long form of all the supported command line options are
        parsed correctly.
        """
        opt = webcmd.WebConfiguration()
        opt.parent = self
        certFile = self.store.filesdir.child('name')
        opt.parseOptions([
                '--http-log', 'http.log', '--hostname', 'example.com',
                '--urchin-key', 'A123'])
        self.assertEquals(opt['http-log'], 'http.log')
        self.assertEquals(opt['hostname'], 'example.com')
        self.assertEquals(opt['urchin-key'], 'A123')


    def test_staticParsing(self):
        """
        Test that the --static option parses arguments of the form
        "url:filename" correctly.
        """
        opt = webcmd.WebConfiguration()
        opt.parent = self
        opt.parseOptions([
                '--static', 'foo:bar',
                '--static', 'quux/fooble:/bar/baz'])
        self.assertEquals(
            opt.staticPaths,
            [('foo', os.path.abspath('bar')),
             ('quux/fooble', '/bar/baz')])


    def test_hostname(self):
        """
        The I{hostname} option changes the C{hostname} attribute of the
        L{SiteConfiguration} object installed on the store.  The hostname
        cannot be set to the empty string.
        """
        opt = webcmd.WebConfiguration()
        opt.parent = self
        opt['hostname'] = 'example.com'
        opt.postOptions()

        opt['hostname'] = ''
        self.assertRaises(UsageError, opt.postOptions)

        self.assertEqual(
            self.store.findUnique(SiteConfiguration).hostname,
            u"example.com")


    def test_urchinKey(self):
        """
        Specifying a Google Analytics key inserts an item into the database
        recording it.
        """
        opt = webcmd.WebConfiguration()
        opt.parent = self
        opt['urchin-key'] = 'A123'
        opt.postOptions()

        self.assertEquals(APIKey.getKeyForAPI(self.store, APIKey.URCHIN).apiKey,
                          u'A123')



# -*- test-case-name: xmantissa.test.test_website -*-
# Copyright 2005 Divmod, Inc.
# See LICENSE file for details

"""
Mantissa web presence.
"""

from zope.interface import implements

from twisted.python.filepath import FilePath
from twisted.internet.defer import maybeDeferred
from twisted.cred.portal import Portal
from twisted.cred.checkers import AllowAnonymousAccess

from nevow.inevow import IRequest, IResource
from nevow.url import URL
from nevow.appserver import NevowSite, NevowRequest
from nevow.rend import NotFound
from nevow.static import File
from nevow.athena import LivePage

from epsilon.structlike import record

from axiom.item import Item
from axiom.attributes import path, text
from axiom.dependency import dependsOn
from axiom.userbase import LoginSystem

from xmantissa.ixmantissa import ISiteURLGenerator, IProtocolFactoryFactory, IOfferingTechnician, ISessionlessSiteRootPlugin
from xmantissa.port import TCPPort, SSLPort
from xmantissa.cachejs import theHashModuleProvider
from xmantissa.websession import PersistentSessionWrapper


class AxiomRequest(NevowRequest):
    def __init__(self, store, *a, **kw):
        NevowRequest.__init__(self, *a, **kw)
        self.store = store

    def process(self, *a, **kw):
        return self.store.transact(NevowRequest.process, self, *a, **kw)



class AxiomSite(NevowSite):
    def __init__(self, store, *a, **kw):
        NevowSite.__init__(self, *a, **kw)
        self.store = store
        self.requestFactory = lambda *a, **kw: AxiomRequest(self.store, *a, **kw)



class SiteConfiguration(Item):
    """
    Configuration object for a Mantissa HTTP server.
    """
    powerupInterfaces = (ISiteURLGenerator, IProtocolFactoryFactory)
    implements(*powerupInterfaces)

    loginSystem = dependsOn(LoginSystem)

    # I don't really want this to have a default value at all, but an Item
    # which can't be instantiated with only a store parameter can't be used as
    # a siteRequirement in an Offering.  See #538 about offering configuration.
    # -exarkun
    hostname = text(
        doc="""
        The primary hostname by which this website will be accessible.  This
        will be superceded by a one-to-many relationship in the future,
        allowing a host to have multiple recognized hostnames.  See #2501.
        """, allowNone=False, default=u"localhost")

    httpLog = path(default=None)


    def _root(self, scheme, hostname, portObj, standardPort):
        # TODO - real unicode support (but punycode is so bad)
        if portObj is None:
            return None

        portNumber = portObj.portNumber
        port = portObj.listeningPort

        if hostname is None:
            hostname = self.hostname
        else:
            hostname = hostname.split(':')[0].encode('ascii')

        if portNumber == 0:
            if port is None:
                return None
            else:
                portNumber = port.getHost().port

        # At some future point, we may want to make pathsegs persistently
        # configurable - perhaps scheme and hostname as well - in order to
        # easily support reverse proxying configurations, particularly where
        # Mantissa is being "mounted" somewhere other than /.  See also rootURL
        # which has some overlap with this method (the difference being
        # universal vs absolute URLs - rootURL may want to call cleartextRoot
        # or encryptedRoot in the future).  See #417 and #2309.
        pathsegs = ['']
        if portNumber != standardPort:
            hostname = '%s:%d' % (hostname, portNumber)
        return URL(scheme, hostname, pathsegs)


    def _getPort(self, portType):
        return self.store.findFirst(
            portType, portType.factory == self, default=None)


    def cleartextRoot(self, hostname=None):
        """
        Return a string representing the HTTP URL which is at the root of this
        site.

        @param hostname: An optional unicode string which, if specified, will
            be used as the hostname in the resulting URL, regardless of the
            C{hostname} attribute of this item.
        """
        return self._root('http', hostname, self._getPort(TCPPort), 80)


    def encryptedRoot(self, hostname=None):
        """
        Return a string representing the HTTPS URL which is at the root of this
        site.

        @param hostname: An optional unicode string which, if specified, will
        be used as the hostname in the resulting URL, regardless of the
        C{hostname} attribute of this item.
        """
        return self._root('https', hostname, self._getPort(SSLPort), 443)


    def rootURL(self, request):
        """
        Return the URL for the root of this website which is appropriate to use
        in links generated in response to the given request.

        @type request: L{twisted.web.http.Request}
        @param request: The request which is being responded to.

        @rtype: L{URL}
        @return: The location at which the root of the resource hierarchy for
            this website is available.
        """
        host = request.getHeader('host') or self.hostname
        if ':' in host:
            host = host.split(':', 1)[0]
        if (host == self.hostname or
            host.startswith('www.') and host[len('www.'):] == self.hostname):
            return URL(scheme='', netloc='', pathsegs=[''])
        else:
            if request.isSecure():
                return self.encryptedRoot(self.hostname)
            else:
                return self.cleartextRoot(self.hostname)


    def getFactory(self):
        """
        Create an L{AxiomSite} which supports authenticated and anonymous
        access.
        """
        checkers = [self.loginSystem, AllowAnonymousAccess()]
        guardedRoot = PersistentSessionWrapper(
            self.store,
            Portal(self.loginSystem, checkers),
            domains=[self.hostname])
        unguardedRoot = UnguardedWrapper(self.store, guardedRoot)
        securingRoot = SecuringWrapper(self, unguardedRoot)
        logPath = None
        if self.httpLog is not None:
            logPath = self.httpLog.path
        return AxiomSite(self.store, securingRoot, logPath=logPath)




class UnguardedWrapper(record('siteStore guardedRoot')):
    """
    Resource which wraps the top of the Mantissa resource hierarchy and adds
    resources which must be available to all clients all the time.

    @ivar siteStore: The site L{Store} for the resource hierarchy being
        wrapped.
    @ivar guardedRoot: The root resource of the hierarchy being wrapped.
    """
    implements(IResource)

    def child_live(self, ctx):
        """
        The 'live' namespace is reserved for Athena LivePages.  By default in
        Athena applications these resources are child resources of whatever URL
        the live page ends up at, but this root URL is provided so that the
        reliable message queuing logic can sidestep all resource traversal, and
        therefore, all database queries.  This is an important optimization,
        since Athena's implementation assumes that HTTP hits to the message
        queue resource are cheap.

        @return: an L{athena.LivePage} instance.
        """
        return LivePage(None, None)


    def child___jsmodule__(self, ignored):
        """
        __jsmodule__ child which provides support for Athena applications to
        use a centralized URL to deploy JavaScript code.
        """
        return theHashModuleProvider


    def child_Mantissa(self, ctx):
        """
        Serve files from C{xmantissa/static/} at the URL C{/Mantissa}.
        """
        # Cheating!  It *looks* like there's an app store, but there isn't
        # really, because this is the One Store To Bind Them All.

        # We shouldn't really cheat here.  It would be better to have one real
        # Mantissa offering that has its static content served up the same way
        # every other offering's content is served.  There's already a
        # /static/mantissa-static/.  This child definition is only still here
        # because some things still reference this URL.  For example,
        # JavaScript files and any CSS file which uses Mantissa content but is
        # from an Offering which does not provide a staticContentPath.
        # See #2469.  -exarkun
        return File(FilePath(__file__).sibling("static").path)


    def child_static(self, context):
        """
        Serve a container page for static content for Mantissa and other
        offerings.
        """
        offeringTech = IOfferingTechnician(self.siteStore)
        installedOfferings = offeringTech.getInstalledOfferings()
        offeringsWithContent = dict([
                (offering.name, offering.staticContentPath)
                for offering
                in installedOfferings.itervalues()
                if offering.staticContentPath])

        # If you wanted to do CSS rewriting for all CSS files served beneath
        # /static/, you could do it by passing a processor for ".css" here.
        # eg:
        #
        # website = IResource(self.store)
        # factory = StylesheetFactory(
        #     offeringsWithContent.keys(), website.rootURL)
        # StaticContent(offeringsWithContent, {
        #               ".css": factory.makeStylesheetResource})
        return StaticContent(offeringsWithContent, {})


    def locateChild(self, context, segments):
        """
        Return a statically defined child or a child defined by a sessionless
        site root plugin or an avatar from guard.
        """
        shortcut = getattr(self, 'child_' + segments[0], None)
        if shortcut:
            res = shortcut(context)
            if res is not None:
                return res, segments[1:]

        req = IRequest(context)
        for plg in self.siteStore.powerupsFor(ISessionlessSiteRootPlugin):
            spr = getattr(plg, 'sessionlessProduceResource', None)
            if spr is not None:
                childAndSegments = spr(req, segments)
            else:
                childAndSegments = plg.resourceFactory(segments)
            if childAndSegments is not None:
                return childAndSegments

        return self.guardedRoot.locateChild(context, segments)



class SecuringWrapper(record('urlGenerator wrappedResource')):
    """
    Resource wrapper which enforces HTTPS under certain circumstances.

    If a child resource ultimately obtained from this wrapper is going to be
    rendered, if it has a C{needsSecure} attribute set to a true value and
    there is an HTTPS server available and the request was made over HTTP, the
    client will be redirected to an HTTPS location first.  Otherwise the
    resource will be rendered as usual.

    @ivar urlGenerator: The L{ISiteURLGenerator} which will provide an HTTPS
        URL.
    @ivar wrappedResource: The resource which will be used to locate children
        or which will be rendered.
    """
    implements(IResource)

    def locateChild(self, context, segments):
        """
        Unwrap the wrapped resource if HTTPS is already being used, otherwise
        wrap it in a helper which will preserve the wrapping all the way down
        to the final resource.
        """
        request = IRequest(context)
        if request.isSecure():
            return self.wrappedResource, segments
        return _SecureWrapper(self.urlGenerator, self.wrappedResource), segments


    def renderHTTP(self, context):
        """
        Render the wrapped resource if HTTPS is already being used, otherwise
        invoke a helper which may generate a redirect.
        """
        request = IRequest(context)
        if request.isSecure():
            renderer = self.wrappedResource
        else:
            renderer = _SecureWrapper(self.urlGenerator, self.wrappedResource)
        return renderer.renderHTTP(context)



class _SecureWrapper(record('urlGenerator wrappedResource')):
    """
    Helper class for L{SecuringWrapper} which preserves wrapping to the
    ultimate resource and which implements HTTPS redirect logic if necessary.

    @ivar urlGenerator: The L{ISiteURLGenerator} which will provide an HTTPS
        URL.
    @ivar wrappedResource: The resource which will be used to locate children
        or which will be rendered.
    """
    implements(IResource)

    def __init__(self, *a, **k):
        super(_SecureWrapper, self).__init__(*a, **k)
        self.wrappedResource = IResource(self.wrappedResource)


    def locateChild(self, context, segments):
        """
        Delegate child lookup to the wrapped resource but wrap whatever results
        in another instance of this wrapper.
        """
        childDeferred = maybeDeferred(
            self.wrappedResource.locateChild, context, segments)
        def childLocated((resource, segments)):
            if (resource, segments) == NotFound:
                return NotFound
            return _SecureWrapper(self.urlGenerator, resource), segments
        childDeferred.addCallback(childLocated)
        return childDeferred


    def renderHTTP(self, context):
        """
        Check to see if the wrapped resource wants to be rendered over HTTPS
        and generate a redirect if this is so, if HTTPS is available, and if
        the request is not already over HTTPS.
        """
        if getattr(self.wrappedResource, 'needsSecure', False):
            request = IRequest(context)
            url = self.urlGenerator.encryptedRoot()
            if url is not None:
                for seg in request.prepath:
                    url = url.child(seg)
                return url
        return self.wrappedResource.renderHTTP(context)



class StaticContent(record('staticPaths processors')):
    """
    Parent resource for all static content provided by all installed offerings.

    This resource has a child by the name of each offering which declares a
    static content path which serves that path.

    @ivar staticPaths: A C{dict} mapping offering names to L{FilePath}
        instances for each offering which should be able to publish static
        content.

    @ivar processors: A C{dict} mapping extensions (with leading ".") to
        two-argument callables.  These processors will be attached to the
        L{nevow.static.File} returned by C{locateChild}.
    """
    implements(IResource)

    def locateChild(self, context, segments):
        """
        Find the offering with the name matching the first segment and return a
        L{File} for its I{staticContentPath}.
        """
        name = segments[0]
        try:
            staticContent = self.staticPaths[name]
        except KeyError:
            return NotFound
        else:
            resource = File(staticContent.path)
            resource.processors = self.processors
            return resource, segments[1:]
        return NotFound




"""
Tests for common functionality between L{IWebViewer} implementations.

No runnable tests are currently defined here; the mixin defined here is
imported from test_webapp (for the authenticated view) and test_publicweb (for
the anonymous view).
"""

from axiom.userbase import LoginSystem
from axiom.store import Store

from nevow.rend import Page, NotFound
from nevow.page import deferflatten, Element
from nevow.loaders import stan

from nevow.athena import LivePage
from nevow import tags
from nevow.inevow import IRequest
from nevow.context import WovenContext
from nevow.testutil import FakeRequest

from xmantissa.error import CouldNotLoadFromThemes
from xmantissa.ixmantissa import ISiteURLGenerator
from xmantissa.webtheme import theThemeCache

from xmantissa.test.fakes import (
    FakeLoader, FakeTheme, FakeFragmentModel, FakeLiveElementModel,
    FakeLiveFragmentModel, FakeModel, ResourceViewForFakeModel,
    FakeElementModel, FakeElementModelWithTheme, FakeElementModelWithDocFactory,
    FakeElementModelWithThemeAndDocFactory, ElementViewForFakeModelWithTheme,
    FakeElementModelWithLocateChildView, FakeElementModelWithHead,
    ElementViewForFakeModelWithThemeAndDocFactory, __file__ as fakesfile)

from xmantissa.website import MantissaLivePage
from axiom.plugins.mantissacmd import Mantissa

class WebViewerTestMixin(object):
    """
    Tests for implementations of L{IWebViewer}.
    """

    def setUp(self):
        """
        Create a site store containing an admin user, then do
        implementation-specific setup.
        """
        self.siteStore = Store(filesdir=self.mktemp())
        m = Mantissa()
        m.installSite(self.siteStore, u"localhost", u"", False)
        m.installAdmin(self.siteStore, u'admin', u'localhost', u'asdf')
        self.loginSystem = self.siteStore.findUnique(LoginSystem)
        self.adminStore = self.loginSystem.accountByAddress(
            u'admin', u'localhost').avatars.open()
        self.setupPageFactory()


    def assertPage(self, page):
        """
        Fail if the given object is not a nevow L{rend.Page} (or if it is an
        L{LivePage}).
        """
        self.assertIsInstance(page, Page)
        self.assertNotIsInstance(page, LivePage)


    def assertLivePage(self, page):
        """
        Fail if the given object is not a L{MantissaLivePage}
        """
        self.assertIsInstance(page, MantissaLivePage)


    def test_wrapResource(self):
        """
        An object adaptable to L{IResource} has its IResource adapter returned
        when wrapModel is called with it.
        """
        resourceable = FakeModel()
        result = self.pageFactory.wrapModel(resourceable)
        self.assertIsInstance(result, ResourceViewForFakeModel)


    def test_wrapElement(self):
        """
        An object adaptable to L{INavigableFragment}, whose adapter is an
        L{Element}, is wrapped in a shell page (containing navigation for
        the current user) when wrapModel is called on it.
        """
        elementable = FakeElementModel()
        result = self.pageFactory.wrapModel(elementable)
        self.assertPage(result)
        self.assertEqual(result.fragment.model, elementable)


    def test_wrapFragment(self):
        """
        An object adaptable to L{INavigableFragment}, whose adapter is a
        L{Fragment}, is wrapped in a shell page (containing navigation for
        the current user) when wrapModel is called on it.
        """
        fragmentable = FakeFragmentModel()
        result = self.pageFactory.wrapModel(fragmentable)
        self.assertPage(result)
        self.assertEqual(result.fragment.original, fragmentable)


    def test_wrapLiveElement(self):
        """
        An object adaptable to L{INavigableFragment}, whose adapter is a
        L{LiveElement}, is wrapped in a shell page (containing navigation for
        the current user) when wrapModel is called on it.
        """
        elementable = FakeLiveElementModel()
        result = self.pageFactory.wrapModel(elementable)
        self.assertLivePage(result)
        self.assertEqual(result.fragment.model, elementable)


    def test_wrapLiveFragment(self):
        """
        An object adaptable to L{INavigableFragment}, whose adapter is a
        L{LiveFragment}, is wrapped in a shell page (containing navigation for
        the current user) when wrapModel is called on it.
        """
        fragmentable = FakeLiveFragmentModel()
        result = self.assertWarns(
            DeprecationWarning,
            '[v0.10] LiveFragment has been superceded by LiveElement.',
            fakesfile,
            self.pageFactory.wrapModel, fragmentable)
        self.assertLivePage(result)
        self.assertEqual(result.fragment.original, fragmentable)


    def stubThemeList(self, themes):
        """
        Replace the global themes list with the given list of themes for the
        duration of this test.  This relies on the implementation details of
        the global theme cache; please update if the theme cache implementation
        changes.
        """
        theThemeCache._getInstalledThemesCache[self.siteStore] = themes
        self.addCleanup(theThemeCache.emptyCache)


    def test_docFactoryFromFragmentName(self):
        """
        If the L{INavigableFragment} provider provides a C{fragmentName}
        attribute, it should be looked up via the theme system and assigned to
        the C{docFactory} attribute of the L{INavigableFragment} provider.
        """
        fakeDocFactory = FakeLoader('fake')
        self.stubThemeList([FakeTheme('theme',
                    {ElementViewForFakeModelWithTheme.fragmentName:
                         fakeDocFactory})])
        elementable = FakeElementModelWithTheme()
        result = self.pageFactory.wrapModel(elementable)
        self.assertEqual(result.fragment.docFactory, fakeDocFactory)


    def test_dontThrowOutTheBabyWithTheBathwater(self):
        """
        If the L{INavigableFragment} provider provides a C{docFactory}
        attribute, neither lack of a C{fragmentName} attribute, a
        C{fragmentName} that is None, nor a C{fragmentName} for which there is
        no loader in any theme, should change its C{docFactory} to C{None}.
        """
        fakeDocFactory = FakeLoader('fake')
        self.stubThemeList([FakeTheme('theme', {})])
        model = FakeElementModelWithDocFactory(fakeDocFactory)
        modelWithFragmentName = FakeElementModelWithThemeAndDocFactory('wrong',
                                                                       fakeDocFactory)

        result = self.pageFactory.wrapModel(model)
        self.assertEqual(result.fragment.docFactory, fakeDocFactory)

        result = self.pageFactory.wrapModel(modelWithFragmentName)
        self.assertEqual(result.fragment.docFactory, fakeDocFactory)


    def test_noLoaderAnywhere(self):
        """
        If the L{INavigableFragment}provider provides no C{docFactory} or
        C{fragmentName} attribute, then L{_AuthenticatedWebViewer.wrapModel}
        should raise L{CouldNotLoadFromThemes}.
        """
        model = FakeElementModelWithThemeAndDocFactory(None, None)
        aFakeTheme = FakeTheme('fake', {})
        self.stubThemeList([aFakeTheme])
        exc = self.assertRaises(
            CouldNotLoadFromThemes, self.pageFactory.wrapModel, model)
        self.assertIsInstance(exc.element, ElementViewForFakeModelWithThemeAndDocFactory)
        self.assertIdentical(exc.element.model, model)
        self.assertEquals(exc.themes, [aFakeTheme])
        self.assertEquals(
            repr(exc),
            'CouldNotLoadFromThemes: %r (fragment name %r) has no docFactory. Searched these themes: %r' %
            (exc.element, exc.element.fragmentName, [aFakeTheme]))


    def test_navFragmentHasLocateChild(self, beLive=False):
        """
        The L{IResource} returned by L{IWebViewer.wrapModel} must have a
        C{locateChild} which forwards to its wrapped L{INavigableFragment}.
        """
        expectChild = object()
        elementable = FakeElementModelWithLocateChildView([expectChild], beLive=beLive)
        result = self.pageFactory.wrapModel(elementable)
        resultChild = result.locateChild(FakeRequest(), [''])
        self.assertIdentical(resultChild, expectChild)


    def test_navFragmentHasLiveLocateChild(self):
        """
        The same as L{test_navFragmentHasLiveLocateChild}, but for the Athena
        implementation of the resource adapters.
        """
        self.test_navFragmentHasLocateChild(True)


    def test_navFragmentHasNoChildMethods(self, beLive=False):
        """
        The L{IResource} returned by L{IWebViewer.wrapModel} must have a
        C{locateChild} which returns NotFound in the case where its wrapped
        L{INavigableFragment} does not have a C{locateChild} method.
        """
        if beLive:
            elementable = FakeLiveElementModel()
        else:
            elementable = FakeElementModel()
        result = self.pageFactory.wrapModel(elementable)
        resultChild = result.locateChild(FakeRequest(), [''])
        self.assertEqual(resultChild, NotFound)


    def test_navFragmentHasNoLiveChildMethods(self):
        """
        The same as L{test_navFragmentHasNoChildMethods} but for the Athena
        implementation of the resource adapters.
        """
        self.test_navFragmentHasNoChildMethods(True)


    def _renderHead(self, result):
        """
        Go through all the gyrations necessary to get the head contents
        produced by the page rendering process.
        """
        site = ISiteURLGenerator(self.siteStore)
        t = tags.invisible()
        ctx = WovenContext(tag=t)
        req = FakeRequest()
        ctx.remember(req, IRequest)
        expected = [th.head(req, site)
                    for th in self.pageFactory._preferredThemes()]
        head = result.render_head(ctx, None)
        return t, expected


    def test_renderHeadNoHeadMethod(self):
        """
        Rendering the <head> tag in the shell template for a non-athena
        L{INavigableFragment} without a C{head()} method should result in each
        theme adding its external content (stylesheets, etc) to the <head>.
        """
        elementable = FakeElementModel()
        result = self.pageFactory.wrapModel(elementable)
        t, expected = self._renderHead(result)
        self.assertStanEqual(list(t.children), expected)


    def test_renderHeadWithHeadMethod(self):
        """
        Rendering the <head> tag in the shell template for a non-athena
        L{INavigableFragment} with a C{head()} method should result in each
        theme adding its external content (stylesheets, etc) to the <head>,
        followed by the contents provided by the L{INavigableFragment}.
        """
        h = tags.div["some stuff"]
        model = FakeElementModelWithHead(h)
        result = self.pageFactory.wrapModel(model)
        t, expected = self._renderHead(result)
        self.assertStanEqual(list(t.children), expected + [h])


    def _instaString(self, x):
        """
        Convert a new-style-renderable thing into a string, raising an
        exception if any deferred rendering took place.
        """
        e = Element(docFactory=stan(x))
        l = []
        d = deferflatten(None, e, False, False, l.append)
        return ''.join(l)


    def assertStanEqual(self, stan1, stan2):
        """
        Assert that two chunks of Nevow-renderable stuff will result in the
        same XML.
        """
        a = self._instaString(stan1)
        b = self._instaString(stan2)
        self.assertEqual(a, b)



import sha

from twisted.trial.unittest import TestCase
from twisted.python.filepath import FilePath

from nevow.inevow import IRequest
from nevow.context import WovenContext
from nevow.testutil import FakeRequest

from xmantissa.cachejs import HashedJSModuleProvider, CachedJSModule


class JSCachingTestCase(TestCase):
    """
    Tests for L{xmantissa.cachejs}.
    """
    hostname = 'test-mantissa-js-caching.example.com'

    def setUp(self):
        """
        Create a L{HashedJSModuleProvider} and a dummy module.
        """
        self.MODULE_NAME = 'Dummy.Module'
        self.MODULE_CONTENT = '/* Hello, world. /*\n'
        self.moduleFile = self.mktemp()
        fObj = file(self.moduleFile, 'w')
        fObj.write(self.MODULE_CONTENT)
        fObj.close()
        m = HashedJSModuleProvider()
        self.moduleProvider = m
        self._wasModified = CachedJSModule.wasModified.im_func
        self.callsToWasModified = 0
        def countCalls(other):
            self.callsToWasModified += 1
            return self._wasModified(other)
        CachedJSModule.wasModified = countCalls


    def tearDown(self):
        """
        put L{CachedJSModule} back the way we found it
        """
        CachedJSModule.wasModified = self._wasModified


    def _render(self, resource):
        """
        Test helper which tries to render the given resource.
        """
        ctx = WovenContext()
        req = FakeRequest(headers={'host': self.hostname})
        ctx.remember(req, IRequest)
        return req, resource.renderHTTP(ctx)


    def test_hashExpiry(self):
        """
        L{HashedJSModuleProvider.resourceFactory} should return a L{static.Data}
        with an C{expires} value far in the future.
        """
        self.moduleProvider.moduleCache[self.MODULE_NAME] = CachedJSModule(
            self.MODULE_NAME, FilePath(self.moduleFile))
        d, segs = self.moduleProvider.locateChild(None,
                                     [sha.new(self.MODULE_CONTENT).hexdigest(),
                                      self.MODULE_NAME])
        self.assertEqual([], segs)
        d.time = lambda: 12345
        req, result = self._render(d)
        self.assertEquals(
            req.headers['expires'],
            'Tue, 31 Dec 1974 03:25:45 GMT')
        self.assertEquals(
            result,
            '/* Hello, world. /*\n')


    def test_getModule(self):
        """
        L{HashedJSModuleProvider.getModule} should only load modules once;
        subsequent calls should return the cached module object.
        """
        module = self.moduleProvider.getModule("Mantissa.Test.Dummy")
        self.failUnlessIdentical(module, self.moduleProvider.getModule(
            "Mantissa.Test.Dummy"))


    def test_dontStat(self):
        """
        L{HashedJSModuleProvider.getModule} shouldn't hit the disk more than
        once per module.
        """
        module1 = self.moduleProvider.getModule("Mantissa.Test.Dummy")
        module2 = self.moduleProvider.getModule("Mantissa.Test.Dummy")
        self.assertEqual(self.callsToWasModified, 1)




from twisted.python.filepath import FilePath
from twisted.protocols.amp import IBoxReceiver
from twisted.conch.interfaces import IConchUser

from nevow.inevow import IResource

from xmantissa.ixmantissa import ISiteURLGenerator
from xmantissa import offering
from xmantissa.web import SiteConfiguration
from xmantissa.webtheme import MantissaTheme
from xmantissa.publicweb import AnonymousSite
from xmantissa.ampserver import AMPConfiguration, AMPAvatar, EchoFactory
from xmantissa.terminal import SecureShellConfiguration
import xmantissa

baseOffering = offering.Offering(
    name=u'mantissa-base',
    description=u'Basic Mantissa functionality',
    siteRequirements=[
        (ISiteURLGenerator, SiteConfiguration),
        (IResource, AnonymousSite),
        (None, SecureShellConfiguration)],
    appPowerups=(),
    installablePowerups=(),
    loginInterfaces = [
        (IResource, "HTTP logins"),
        (IConchUser, "SSH logins")],
    # priority should be 0 for pretty much any other theme.  'base' is the theme
    # that all other themes should use as a reference for what elements are
    # required.
    themes=(MantissaTheme('base', 1),),
    staticContentPath=FilePath(xmantissa.__file__).sibling('static'),
    version=xmantissa.version)

# XXX This should be part of baseOffering, but because there is no
# functionality for upgrading installed offering state, doing so would create a
# class of databases which thought they had amp installed but didn't really.
# See #2723.
ampOffering = offering.Offering(
    name=u'mantissa-amp',
    description=u'Extra AMP-related Mantissa functionality',
    siteRequirements=[
        (None, AMPConfiguration),
        ],
    appPowerups=[],
    installablePowerups=[
        ("AMP Access", "Allows logins over AMP", AMPAvatar),
        ("AMP Echo Protocol",
         "Dead-simple AMP echoer, potentially useful for testing connections "
         "to a Mantissa AMP server or otherwise providing an example of the "
         "AMP functionality.",
         EchoFactory),
        ],
    loginInterfaces=[
        (IBoxReceiver, "AMP logins"),
        ],
    themes=[],
    staticContentPath=None,
    version=xmantissa.version)




from axiom.test.historic.stubloader import StubbedTest
from axiom.batch import processor
from axiom.iaxiom import REMOTE

from xmantissa.fulltext import HypeIndexer, XapianIndexer, PyLuceneIndexer
from xmantissa.test.test_fulltext import FakeMessageSource

class RemoteIndexerTestCase(StubbedTest):
    """
    Test that the upgrade from 2 to 3 does not drop any information and that it
    resets the indexer so that the changes to the indexing behavior gets
    applied to all old messages.
    """
    def setUp(self):
        # Grab the FakeMessageSource and hold a reference to it so its
        # in-memory attributes stick around long enough for us to make some
        # assertions about them.
        result = StubbedTest.setUp(self)
        self.messageSource = self.store.findUnique(FakeMessageSource)
        return result

    def _test(self, indexerClass):
        indexer = self.store.findUnique(indexerClass)
        self.assertEquals(
            [self.messageSource],
            list(indexer.getSources()))
        # Make sure it got reset
        self.assertIn((indexer, REMOTE), self.messageSource.added)
        self.assertIn(indexer, self.messageSource.removed)

    def testUpgradeHype(self):
        return self._test(HypeIndexer)


    def testUpgradeXapian(self):
        return self._test(XapianIndexer)


    def testUpgradePyLucene(self):
        return self._test(PyLuceneIndexer)



from zope.interface import implements, Interface

from twisted.trial.unittest import TestCase
from twisted.python.reflect import qual

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer

from xmantissa.product import Installation, ProductConfiguration, Product, ProductFragment

class IFoo(Interface):
    pass

class Foo(Item):
    implements(IFoo)
    powerupInterfaces = (IFoo,)
    attr = integer()

class IBaz(Interface):
    pass

class Baz(Item):
    implements(IBaz)
    powerupInterfaces = (IBaz,)
    attr = integer()

class ProductTest(TestCase):
    def setUp(self):
        """
        Create a pseudo site store and a pseudo user store in it.
        """
        self.siteStore = Store()
        self.userStore = Store()
        self.userStore.parent = self.siteStore


    def test_product(self):
        self.product = Product(store=self.siteStore)
        self.product.types = [
            n.decode('ascii') for n in [qual(Foo), qual(Baz)]]
        self.product.installProductOn(self.userStore)
        i = self.userStore.findUnique(Installation)
        self.assertEqual(i.types, self.product.types)


    def test_createProduct(self):
        """
        Verify that L{ProductConfiguration.createProduct} creates a
        correctly configured L{Product} and returns it.
        """
        conf = ProductConfiguration(store=self.userStore)
        product = conf.createProduct([Foo, Baz])
        self.assertEqual(product.types, [qual(Foo), qual(Baz)])



class InstallationTest(TestCase):

    def setUp(self):
        self.s = Store()
        self.product = Product()
        self.product.types = [n.decode('ascii') for n in [qual(Foo), qual(Baz)]]
        self.product.installProductOn(self.s)
        self.i = self.s.findUnique(Installation)


    def test_install(self):
        """
        Ensure that Installation installs instances of the types it is created with.
        """
        self.assertNotEqual(IFoo(self.s, None), None)
        self.assertNotEqual(IBaz(self.s, None), None)
        self.assertEqual(list(self.i.items), [self.s.findUnique(t) for t in [Foo, Baz]])


    def test_uninstall(self):
        """
        Ensure that Installation properly uninstalls all of the items it controls.
        """
        self.product.removeProductFrom(self.s)
        self.assertEqual(IFoo(self.s, None), None)
        self.assertEqual(IBaz(self.s, None), None)
        self.assertEqual(list(self.s.query(Installation)), [])



class StubProductConfiguration(object):
    """
    Stub implementation of L{ProductConfiguration} for testing purposes.

    @ivar createdProducts: A list containing all powerups lists passed to
    C{createProduct}.
    """
    def __init__(self, createdProducts):
        self.createdProducts = createdProducts


    def createProduct(self, powerups):
        self.createdProducts.append(powerups)



class ViewTests(TestCase):
    """
    Tests for L{ProductFragment}.
    """
    def test_coerceProduct(self):
        """
        Verify that L{ProductFragment.coerceProduct} calls C{createProduct} on
        the object it wraps and passes its arguments through.
        """
        createdProducts = []
        fragment = ProductFragment(StubProductConfiguration(createdProducts))
        fragment.coerceProduct(foo=u'bar', baz=u'quux')
        self.assertEqual(createdProducts, [[u'bar', u'quux']])


    def test_coerceProductReturn(self):
        """
        Verify that L{ProductFragment.coerceProduct} returns a string
        indicating success.
        """
        createdProducts = []
        fragment = ProductFragment(StubProductConfiguration(createdProducts))
        result = fragment.coerceProduct()
        self.assertEqual(result, u'Created.')




from twisted.python.filepath import FilePath

from axiom.test.historic.stubloader import saveStub
from xmantissa.people import Mugshot, Person

def createDatabase(s):
    imgfile = FilePath(__file__).parent().parent().child('resources').child('square.png')
    outfile = s.newFile('the-image')
    outfile.write(imgfile.getContent())
    outfile.close()

    Mugshot(store=s,
            person=Person(store=s, name=u'Bob'),
            body=outfile.finalpath,
            type=u'image/png')

if __name__ == '__main__':
    saveStub(createDatabase, 7671)



from xmantissa import people, offering, website

peopleOffering = offering.Offering(
    name=u'People',
    description=u'Basic organizer and addressbook support.',

    siteRequirements=((None, website.WebSite),),
    appPowerups=(),
    installablePowerups = [("People", "Organizer and Address Book", people.AddPerson)],
    loginInterfaces=(),
    themes=())




# -*- test-case-name: xmantissa.test.historic.test_messagequeue1to2 -*-

from axiom.test.historic.stubloader import saveStub

from axiom.dependency import installOn

from xmantissa.interstore import MessageQueue

MESSAGE_COUNT = 17

def createDatabase(store):
    installOn(MessageQueue(store=store, messageCounter=MESSAGE_COUNT), store)

if __name__ == '__main__':
    saveStub(createDatabase, 17606)



# -*- test-case-name: xmantissa.test.test_fulltext -*-

"""
General functionality re-usable by various concrete fulltext indexing systems.
"""

import atexit, os, weakref, warnings

from zope.interface import implements

from twisted.python import log, reflect
from twisted.internet import defer

from epsilon.structlike import record
from epsilon.view import SlicedView

from axiom import item, attributes, iaxiom, batch
from axiom.upgrade import registerUpgrader, registerAttributeCopyingUpgrader
from axiom.store import Store, AttributeQuery
from axiom.attributes import AttributeValueComparison, SimpleOrdering

from xmantissa import ixmantissa

HYPE_INDEX_DIR = u'hype.index'
XAPIAN_INDEX_DIR = u'xap.index'
LUCENE_INDEX_DIR = u'lucene.index'

VERBOSE = True

class IndexCorrupt(Exception):
    """
    An attempt was made to open an index which has had unrecoverable data
    corruption.
    """



class _IndexerInputSource(item.Item):
    """
    Tracks L{IBatchProcessor}s which have had an indexer added to them as a
    listener.
    """
    indexer = attributes.reference(doc="""
    The indexer item with which this input source is associated.
    """, whenDeleted=attributes.reference.CASCADE)

    source = attributes.reference(doc="""
    The L{IBatchProcessor} which acts as the input source.
    """, whenDeleted=attributes.reference.CASCADE)



class _RemoveDocument(item.Item):
    """
    Tracks a document deletion which should occur before the next search is
    performed.
    """
    indexer = attributes.reference(doc="""
    The indexer item with which this deletion is associated.
    """, whenDeleted=attributes.reference.CASCADE)

    documentIdentifier = attributes.bytes(doc="""
    The identifier, as returned by L{IFulltextIndexable.uniqueIdentifier},
    for the document which should be removed from the index.
    """, allowNone=False)



class RemoteIndexer(object):
    """
    Implements most of a full-text indexer.

    This uses L{axiom.batch} to perform indexing out of process and presents an
    asynchronous interface to in-process searching of that indexing.
    """
    implements(iaxiom.IReliableListener, ixmantissa.ISearchProvider, ixmantissa.IFulltextIndexer)


    def installOn(self, other):
        super(RemoteIndexer, self).installOn(other)
        other.powerUp(self, ixmantissa.IFulltextIndexer)


    def openReadIndex(self):
        """
        Return an object usable to search this index.

        Subclasses should implement this.
        """
        raise NotImplementedError


    def openWriteIndex(self):
        """
        Return an object usable to add documents to this index.

        Subclasses should implement this.
        """
        raise NotImplementedError


    def __finalizer__(self):
        d = self.__dict__
        id = self.storeID
        s = self.store
        def finalize():
            idx = d.get('_index', None)
            if idx is not None:
                if VERBOSE:
                    log.msg("Closing %r from finalizer of %s/%d" % (idx, s, id))
                idx.close()
        return finalize


    def activate(self):
        assert not hasattr(self, '_index')
        self._index = None
        if VERBOSE:
            log.msg("Activating %s/%d with null index" % (self.store, self.storeID))


    def addSource(self, itemSource):
        """
        Add the given L{IBatchProcessor} as a source of input for this indexer.
        """
        _IndexerInputSource(store=self.store, indexer=self, source=itemSource)
        itemSource.addReliableListener(self, style=iaxiom.REMOTE)


    def getSources(self):
        return self.store.query(_IndexerInputSource, _IndexerInputSource.indexer == self).getColumn("source")


    def reset(self):
        """
        Process everything all over again.
        """
        self.indexCount = 0
        indexDir = self.store.newDirectory(self.indexDirectory)
        if indexDir.exists():
            indexDir.remove()
        for src in self.getSources():
            src.removeReliableListener(self)
            src.addReliableListener(self, style=iaxiom.REMOTE)


    def _closeIndex(self):
        if VERBOSE:
            log.msg("%s/%d closing index" % (self.store, self.storeID))
        if self._index is not None:
            if VERBOSE:
                log.msg("%s/%d *really* closing index" % (self.store, self.storeID))
            self._index.close()
            self._index = None


    # IFulltextIndexer
    def add(self, item):
        if self._index is None:
            try:
                self._index = self.openWriteIndex()
            except IndexCorrupt:
                self.reset()
                return

            if VERBOSE:
                log.msg("Opened %s %s/%d for writing" % (self._index, self.store, self.storeID))

        if VERBOSE:
            log.msg("%s/%d indexing document" % (self.store, self.storeID))
        self._index.add(ixmantissa.IFulltextIndexable(item))
        self.indexCount += 1


    def remove(self, item):
        identifier = ixmantissa.IFulltextIndexable(item).uniqueIdentifier()
        if VERBOSE:
            log.msg("%s/%d scheduling %r for removal." % (self.store, self.storeID, identifier))
        _RemoveDocument(store=self.store,
                        indexer=self,
                        documentIdentifier=identifier)



    def _flush(self):
        """
        Deal with pending result-affecting things.

        This should always be called before issuing a search.
        """
        remove = self.store.query(_RemoveDocument)
        documentIdentifiers = list(remove.getColumn("documentIdentifier"))
        if VERBOSE:
            log.msg("%s/%d removing %r" % (self.store, self.storeID, documentIdentifiers))
        reader = self.openReadIndex()
        map(reader.remove, documentIdentifiers)
        reader.close()
        remove.deleteFromStore()


    # IReliableListener
    def suspend(self):
        self._flush() # Make sure any pending deletes are processed.
        if VERBOSE:
            log.msg("%s/%d suspending" % (self.store, self.storeID))
        self._closeIndex()
        return defer.succeed(None)


    def resume(self):
        if VERBOSE:
            log.msg("%s/%d resuming" % (self.store, self.storeID))
        return defer.succeed(None)


    def processItem(self, item):
        return self.add(item)


    # ISearchProvider
    def search(self, aString, keywords=None, count=None, offset=0,
               sortAscending=True, retry=3):
        ident = "%s/%d" % (self.store, self.storeID)
        b = iaxiom.IBatchService(self.store)
        if VERBOSE:
            log.msg("%s issuing suspend" % (ident,))
        d = b.suspend(self.storeID)

        def reallySearch(ign):
            if VERBOSE:
                log.msg("%s getting reader index" % (ident,))
            idx = self.openReadIndex()

            if VERBOSE:
                log.msg("%s searching for %s" % (
                    ident, aString.encode('utf-8')))
            results = idx.search(aString, keywords, sortAscending)
            if VERBOSE:
                log.msg("%s found %d results" % (ident, len(results)))

            if count is None:
                end = None
            else:
                end = offset + count

            results = results[offset:end]

            if VERBOSE:
                log.msg("%s sliced from %s to %s, leaving %d results" % (
                        ident, offset, end, len(results)))
            return results

        d.addCallback(reallySearch)

        def resumeIndexing(results):
            if VERBOSE:
                log.msg("%s issuing resume" % (ident,))
            b.resume(self.storeID).addErrback(log.err)
            return results
        d.addBoth(resumeIndexing)

        def searchFailed(err):
            log.msg("Search failed somehow:")
            log.err(err)
            if retry:
                log.msg("Re-issuing search")
                return self.search(aString, keywords, count, offset, retry=retry-1)
            else:
                log.msg("Wow, lots of failures searching.  Giving up and "
                        "returning (probably wrong!) no results to user.")
                return []
        d.addErrback(searchFailed)
        return d



try:
    import hype
except ImportError:
    hype = None

class _HypeIndex(object):
    def __init__(self, index):
        self.index = index
        self.close = index.close


    def add(self, message):
        doc = hype.Document()
        for (k, v) in message.valueParts():
            doc.add_hidden_text(v.encode('utf-8'))
        doc['@uri'] = message.uniqueIdentifier()

        for part in message.textParts():
            doc.add_text(part.encode('utf-8'))

        self.index.put_doc(doc)


    def search(self, term, keywords=None, sortAscending=True):
        return [int(d.uri) for d in self.index.search(term)]



class HypeIndexer(RemoteIndexer, item.Item):

    schemaVersion = 3

    indexCount = attributes.integer(default=0)
    installedOn = attributes.reference()
    indexDirectory = attributes.text(default=HYPE_INDEX_DIR)

    _index = attributes.inmemory()

    if hype is None:
        def openReadIndex(self):
            raise NotImplementedError("hype is unavailable")


        def openWriteIndex(self):
            raise NotImplementedError("hype is unavailable")
    else:
        def openReadIndex(self):
            hypedir = self.store.newDirectory(self.indexDirectory)
            return _HypeIndex(hype.Database(hypedir.path, hype.ESTDBREADER | hype.ESTDBLCKNB | hype.ESTDBCREAT))


        def openWriteIndex(self):
            hypedir = self.store.newDirectory(self.indexDirectory)
            return _HypeIndex(hype.Database(hypedir.path, hype.ESTDBWRITER | hype.ESTDBCREAT))


class XapianIndexer(RemoteIndexer, item.Item):
    """
    The remnants of an indexer based on Xapian (by way of Xapwrap).  This
    indexing back end is no longer supported.  This item remains defined for
    schema compatibility only.  It should be upgraded out of existence
    eventually and then the class deleted.
    """

    schemaVersion = 3

    indexCount = attributes.integer(default=0)
    installedOn = attributes.reference()
    indexDirectory = attributes.text(default=XAPIAN_INDEX_DIR)

    _index = attributes.inmemory()

    def openReadIndex(self):
        raise NotImplementedError("xapian is no longer supported")


    def openWriteIndex(self):
        raise NotImplementedError("xapian is no longer supported")



try:
    import PyLucene
except ImportError:
    PyLucene = None


_hitsWrapperWeakrefs = weakref.WeakKeyDictionary()

class _PyLuceneHitsWrapper(record('index hits')):
    """
    Container for a C{Hits} instance and the L{_PyLuceneIndex} from which it
    came.  This gives the C{Hits} instance a sequence-like interface and when a
    _PyLuceneHitsWrapper is garbage collected, it closes the L{_PyLuceneIndex}
    it has a reference to.
    """
    def __init__(self, *a, **kw):
        super(_PyLuceneHitsWrapper, self).__init__(*a, **kw)

        def close(ref, index=self.index):
            log.msg("Hits wrapper expiring, closing index.")
            index.close()
        _hitsWrapperWeakrefs[self] = weakref.ref(self, close)


    def __len__(self):
        return len(self.hits)


    def __getitem__(self, index):
        """
        Retrieve the storeID field of the requested hit, converting it to an
        integer before returning it.  This handles integer indexes as well as
        slices.
        """
        if isinstance(index, slice):
            return SlicedView(self, index)
        if index >= len(self.hits):
            raise IndexError(index)
        return _PyLuceneHitWrapper(self.hits[index])


class _PyLuceneHitWrapper:
    """
    Wrapper around a single pylucene hit

    @ivar keywordParts: dictionary mapping keyword names to values.  should be the
    same as the result of calling the L{IFulltextIndexable.keywordParts} on
    the item corresponding to the hit
    @ivar documentType: the document type.  return value of
    L{IFulltextIndexable.documentType} called on the item corresponding to the
    hit
    @ivar uniqueIdentifier: an opaque unique identifier.  return value of
    L{IFulltextIndexable.uniqueIdentifier} called on the item corresponding to
    the hit
    @ivar sortKey: the key to sort on.  return value of
    L{IFulltextIndexable.sortKey} called on the item corresponding to the hit
    """
    def __init__(self, hit):
        self.keywordParts = self._getKeywords(hit)
        self.documentType = hit['documentType']
        self.uniqueIdentifier = hit['storeID']
        self.sortKey = hit['sortKey']

    def _getKeywords(self, hit):
        keywords = {}
        systemKeywords = set(('storeID', 'documentType', 'sortKey'))
        for field in hit.fields():
            if field.name() not in systemKeywords:
                keywords[field.name()] = hit[field.name()]
        return keywords

    def __int__(self):
        warnings.warn(
            '_PyLuceneHitWrapper is not an integer',
            DeprecationWarning)
        return int(self.uniqueIdentifier)

    def __cmp__(self, other):
        warnings.warn(
            '_PyLuceneHitWrapper is not an integer',
            DeprecationWarning)
        return int(self).__cmp__(other)


class _PyLuceneBase(object):
    closed = False

    def __init__(self, fsdir, analyzer):
        _closeObjects.append(self)
        self.fsdir = fsdir
        self.analyzer = analyzer


    def close(self):
        if not self.closed:
            self._reallyClose()
            self.fsdir.close()
        self.closed = True
        try:
            _closeObjects.remove(self)
        except ValueError:
            pass



_closeObjects = []
def _closeIndexes():
    """
    Helper for _PyLuceneIndex to make sure FSDirectory and IndexWriter
    instances always get closed.  This gets registered with atexit and
    closes any _PyLuceneIndex objects still in _closeObjects when it gets
    run.
    """
    while _closeObjects:
        _closeObjects[-1].close()
atexit.register(_closeIndexes)



class _PyLuceneReader(_PyLuceneBase):
    """
    Searches and deletes from a Lucene index.
    """
    def __init__(self, fsdir, analyzer, reader, searcher):
        self.reader = reader
        self.searcher = searcher
        super(_PyLuceneReader, self).__init__(fsdir, analyzer)


    def _reallyClose(self):
        self.reader.close()
        self.searcher.close()


    def remove(self, documentIdentifier):
        self.reader.deleteDocuments(
            PyLucene.Term('storeID', documentIdentifier))


    def search(self, phrase, keywords=None, sortAscending=True):
        if not phrase and not keywords:
            return []

        # XXX Colons in phrase will screw stuff up.  Can they be quoted or
        # escaped somehow?  Probably by using a different QueryParser.
        if keywords:
            fieldPhrase = u' '.join(u':'.join((k, v)) for (k, v) in keywords.iteritems())
            if phrase:
                phrase = phrase + u' ' + fieldPhrase
            else:
                phrase = fieldPhrase
        phrase = phrase.translate({ord(u'@'): u' ', ord(u'-'): u' ',
                                   ord(u'.'): u' '})
        qp = PyLucene.QueryParser('text', self.analyzer)
        qp.setDefaultOperator(qp.Operator.AND)
        query = qp.parseQuery(phrase)

        sort = PyLucene.Sort(PyLucene.SortField('sortKey', not sortAscending))

        try:
            hits = self.searcher.search(query, sort)
        except PyLucene.JavaError, err:
            if 'no terms in field sortKey' in str(err):
                hits = []
            else:
                raise
        return _PyLuceneHitsWrapper(self, hits)



class _PyLuceneWriter(_PyLuceneBase):
    """
    Adds documents to a Lucene index.
    """
    def __init__(self, fsdir, analyzer, writer):
        self.writer = writer
        super(_PyLuceneWriter, self).__init__(fsdir, analyzer)


    def _reallyClose(self):
        self.writer.close()


    def add(self, message):
        doc = PyLucene.Document()
        for part in message.textParts():
            doc.add(
                PyLucene.Field('text',
                               part.translate({
                ord(u'@'): u' ', ord(u'-'): u' ',
                ord(u'.'): u' '}).encode('utf-8'),
                               PyLucene.Field.Store.NO,
                               PyLucene.Field.Index.TOKENIZED))

        for (k, v) in message.keywordParts().iteritems():
            doc.add(
                PyLucene.Field(k, v.translate({
                ord(u'@'): u' ', ord(u'-'): u' ',
                ord(u'.'): u' '}).encode('utf-8'),
                            PyLucene.Field.Store.YES,
                            PyLucene.Field.Index.TOKENIZED))
        doc.add(
            PyLucene.Field('documentType', message.documentType(),
                           PyLucene.Field.Store.YES,
                           PyLucene.Field.Index.TOKENIZED))

        doc.add(
            PyLucene.Field('storeID',
                           message.uniqueIdentifier(),
                           PyLucene.Field.Store.YES,
                           PyLucene.Field.Index.UN_TOKENIZED))
        doc.add(
            PyLucene.Field('sortKey',
                           message.sortKey(),
                           PyLucene.Field.Store.YES,
                           PyLucene.Field.Index.UN_TOKENIZED))
        # Deprecated. use Field(name, value, Field.Store.YES, Field.Index.UN_TOKENIZED) instead

        self.writer.addDocument(doc)



class PyLuceneIndexer(RemoteIndexer, item.Item):

    schemaVersion = 5

    indexCount = attributes.integer(default=0)
    installedOn = attributes.reference()
    indexDirectory = attributes.text(default=LUCENE_INDEX_DIR)

    _index = attributes.inmemory()
    _lockfile = attributes.inmemory()


    def reset(self):
        """
        In addition to the behavior of the superclass, delete any dangling
        lockfiles which may prevent this index from being opened.  With the
        tested version of PyLucene (something pre-2.0), this appears to not
        actually be necessary: deleting the entire index directory but
        leaving the lockfile in place seems to still allow the index to be
        recreated (perhaps because when the directory does not exist, we
        pass True as the create flag when opening the FSDirectory, I am
        uncertain).  Nevertheless, do this anyway for now.
        """
        RemoteIndexer.reset(self)
        if hasattr(self, '_lockfile'):
            os.remove(self._lockfile)
            del self._lockfile


    def _analyzer(self):
        return PyLucene.StandardAnalyzer([])


    if PyLucene is None:
        def openReadIndex(self):
            raise NotImplementedError("PyLucene is unavailable")


        def openWriteIndex(self):
            raise NotImplementedError("PyLucene is unavailable")
    else:
        def openReadIndex(self):
            luceneDir = self.store.newDirectory(self.indexDirectory)


            if not luceneDir.exists():
                self.openWriteIndex().close()

            fsdir = PyLucene.FSDirectory.getDirectory(luceneDir.path, False)
            try:
                searcher = PyLucene.IndexSearcher(fsdir)
            except PyLucene.JavaError, e:
                raise IndexCorrupt()
            try:
                reader = PyLucene.IndexReader.open(fsdir)
            except PyLucene.JavaError, e:
                raise IndexCorrupt()
            return _PyLuceneReader(fsdir, self._analyzer(), reader, searcher)


        def openWriteIndex(self):
            luceneDir = self.store.newDirectory(self.indexDirectory)

            create = not luceneDir.exists()

            analyzer = self._analyzer()

            fsdir = PyLucene.FSDirectory.getDirectory(luceneDir.path, create)
            try:
                writer = PyLucene.IndexWriter(fsdir, analyzer, create)
            except PyLucene.JavaError, e:
                lockTimeout = u'Lock obtain timed out: Lock@'
                msg = e.getJavaException().getMessage()
                if msg.startswith(lockTimeout):
                    self._lockfile = msg[len(lockTimeout):]
                raise IndexCorrupt()
            return _PyLuceneWriter(fsdir, analyzer, writer)



def remoteIndexer1to2(oldIndexer):
    """
    Previously external application code was responsible for adding a
    RemoteListener to a batch work source as a reliable listener.  This
    precluded the possibility of the RemoteListener resetting itself
    unilaterally.  With version 2, RemoteListener takes control of adding
    itself as a reliable listener and keeps track of the sources with which it
    is associated.  This upgrader creates that tracking state.
    """
    newIndexer = oldIndexer.upgradeVersion(
        oldIndexer.typeName, 1, 2,
        indexCount=oldIndexer.indexCount,
        installedOn=oldIndexer.installedOn,
        indexDirectory=oldIndexer.indexDirectory)

    listeners = newIndexer.store.query(
        batch._ReliableListener,
        batch._ReliableListener.listener == newIndexer)

    for listener in listeners:
        _IndexerInputSource(
            store=newIndexer.store,
            indexer=newIndexer,
            source=listener.processor)

    return newIndexer

def remoteIndexer2to3(oldIndexer):
    """
    The documentType keyword was added to all indexable items.  Indexes need to
    be regenerated for this to take effect.  Also, PyLucene no longer stores
    the text of messages it indexes, so deleting and re-creating the indexes
    will make them much smaller.
    """
    newIndexer = oldIndexer.upgradeVersion(
        oldIndexer.typeName, 2, 3,
        indexCount=oldIndexer.indexCount,
        installedOn=oldIndexer.installedOn,
        indexDirectory=oldIndexer.indexDirectory)
    # the 3->4 upgrader for PyLuceneIndexer calls reset(), so don't do it
    # here.  also, it won't work because it's a DummyItem
    if oldIndexer.typeName != PyLuceneIndexer.typeName:
        newIndexer.reset()
    return newIndexer


def _declareLegacyIndexerItem(typeClass, version):
    item.declareLegacyItem(typeClass.typeName, version,
                           dict(indexCount=attributes.integer(),
                                installedOn=attributes.reference(),
                                indexDirectory=attributes.text()))

for cls in [HypeIndexer, XapianIndexer, PyLuceneIndexer]:
    _declareLegacyIndexerItem(cls, 2)

    registerUpgrader(
        remoteIndexer1to2,
        item.normalize(reflect.qual(cls)),
        1,
        2)
    registerUpgrader(
        remoteIndexer2to3,
        item.normalize(reflect.qual(cls)),
        2,
        3)
del cls

_declareLegacyIndexerItem(PyLuceneIndexer, 3)

# Copy attributes.  Rely on pyLuceneIndexer4to5 to reset the index due to
# sorting changes.
registerAttributeCopyingUpgrader(PyLuceneIndexer, 3, 4)

_declareLegacyIndexerItem(PyLuceneIndexer, 4)

def pyLuceneIndexer4to5(old):
    """
    Copy attributes, reset index due because information about deleted
    documents has been lost, and power up for IFulltextIndexer so other code
    can find this item.
    """
    new = old.upgradeVersion(PyLuceneIndexer.typeName, 4, 5,
                             indexCount=old.indexCount,
                             installedOn=old.installedOn,
                             indexDirectory=old.indexDirectory)
    new.reset()
    new.store.powerUp(new, ixmantissa.IFulltextIndexer)
    return new

registerUpgrader(pyLuceneIndexer4to5, PyLuceneIndexer.typeName, 4, 5)



class _SQLiteResultWrapper(object):
    """
    Trivial wrapper around SQLite FTS search results.
    """
    def __init__(self, docId):
        self.uniqueIdentifier = docId



class _SQLiteIndex(object):
    """
    FTS3 index interface.
    """

    addSQL = """
    INSERT INTO fts (docid, content) VALUES (?, ?)
    """

    removeSQL = """
    DELETE FROM fts WHERE docid = ?
    """

    searchSQL = """
    SELECT docid
    FROM fts
    WHERE content MATCH ?
    ORDER BY docid %s
    """

    def __init__(self, store):
        self.store = store
        self.close = self.store.close


    def add(self, document):
        """
        Add a document to the database.
        """
        docid = int(document.uniqueIdentifier())
        text = u' '.join(document.textParts())

        self.store.executeSQL(self.addSQL, (docid, text))


    def remove(self, docid):
        """
        Remove a document from the database.
        """
        docid = int(docid)
        self.store.executeSQL(self.removeSQL, (docid,))


    def search(self, term, keywords=None, sortAscending=True):
        """
        Search the database.
        """
        if sortAscending:
            direction = 'ASC'
        else:
            direction = 'DESC'

        return [_SQLiteResultWrapper(r[0]) for r in
                self.store.querySQL(self.searchSQL % (direction,), (term,))]



class SQLiteIndexer(RemoteIndexer, item.Item):
    """
    Indexer implementation using SQLite FTS3.

    XXX: Keywords are currently not supported; see #2877
    """
    indexCount = attributes.integer(default=0)
    indexDirectory = attributes.text(default=u'sqlite.index')

    _index = attributes.inmemory()

    schemaSQL = """
    CREATE VIRTUAL TABLE fts
    USING fts3(content)
    """

    def _getStore(self):
        """
        Get the Store used for FTS.

        If it does not exist, it is created and initialised.
        """
        storeDir = self.store.newDirectory(self.indexDirectory)
        if not storeDir.exists():
            store = Store(storeDir)
            self._initStore(store)
            return store
        else:
            return Store(storeDir)


    def _initStore(self, store):
        """
        Initialise a store for FTS use.
        """
        store.createSQL('CREATE VIRTUAL TABLE fts USING fts3')


    def openReadIndex(self):
        return self.openWriteIndex()


    def openWriteIndex(self):
        return _SQLiteIndex(self._getStore())



# Copyright (c) 2006 Divmod.
# See LICENSE for details.

"""
Runs mantissa javascript tests as part of the mantissa python tests
"""

from nevow.testutil import JavaScriptTestCase

class MantissaJavaScriptTestCase(JavaScriptTestCase):
    """
    Run all the mantissa javascript test
    """

    def test_scrollmodel(self):
        """
        Test the model object which tracks most client-side state for any
        ScrollTable.
        """
        return 'Mantissa.Test.TestScrollModel'


    def test_placeholders(self):
        """
        Test the model objects which track placeholder nodes in the message
        scrolltable.
        """
        return 'Mantissa.Test.TestPlaceholder'


    def test_autocomplete(self):
        """
        Tests the model object which tracks client-side autocomplete state
        """
        return 'Mantissa.Test.TestAutoComplete'


    def test_region(self):
        """
        Test the model objects which track placeholder nodes in the message
        scrolltable.
        """
        return 'Mantissa.Test.TestRegionModel'


    def test_people(self):
        """
        Tests the model objects which deal with the address book and person
        objects.
        """
        return 'Mantissa.Test.TestPeople'


    def test_validate(self):
        """
        Test the class which validates input on the signup page and posts it to
        the server.
        """
        return 'Mantissa.Test.TestValidate'


    def test_liveform(self):
        """
        Test the LiveForm widgets.
        """
        return 'Mantissa.Test.TestLiveForm'


    def test_domReplace(self):
        """
        Test the stuff which replaces things in the DOM.
        """
        return 'Mantissa.Test.TestDOMReplace'


    def test_offering(self):
        """
        Tests for the offering administration interface.
        """
        return 'Mantissa.Test.TestOffering'



from axiom.test.historic.stubloader import saveStub
from xmantissa.settings import Settings

def createDatabase(s):
    Settings(store=s, installedOn=s)

if __name__ == '__main__':
    saveStub(createDatabase, 8528)



import datetime

from nevow.livetrial import testcase
from nevow import loaders, tags, athena

from epsilon import extime

from xmantissa import webadmin


class AdminStatsTestBase(webadmin.AdminStatsFragment):
    docFactory = loaders.stan(tags.div(render=tags.directive('liveElement')))

    def _initializeObserver(self):
        pass


    def getGraphNames(self):
        return [(u"graph1", u"graph 1"), (u"graph2", u"graph 2")]
    athena.expose(getGraphNames)


    def fetchLastHour(self, name):
        t = extime.Time()
        return ([unicode((t + datetime.timedelta(minutes=i)).asHumanly())
                 for i in range(60)],
                [24, 28, 41, 37, 39, 25, 44, 32, 41, 45, 44, 47, 24, 28,
                 29, 49, 43, 56, 28, 35, 66, 43, 72, 65, 62, 56, 84, 52,
                 74, 73, 74, 77, 71, 46, 70, 55, 65, 51, 42, 55, 19, 30,
                 25, 24, 20, 16, 39, 22, 39, 29, 29, 18, 39, 19, 21, 12,
                 25, 25, 25, 29])


    def buildPie(self):
        self.queryStats = {u'beans': 10, u'enchiladas': 27, u'salsa': 3,
                           u'fajitas': 48}
        self.pieSlices()
    athena.expose(buildPie)



class StatsTestCase(testcase.TestCase):
    jsClass = u'Mantissa.Test.StatsTest'
    docFactory = loaders.stan(
        tags.div(render=tags.directive('liveTest'))[
            tags.invisible(render=tags.directive('start'))])


    def render_start(self, ctx, data):
        self.testfragment = AdminStatsTestBase()
        self.testfragment.setFragmentParent(self)
        return self.testfragment


    def run(self):
        self.testfragment.statUpdate(extime.Time(), [(u'graph1', 43)])
        self.testfragment.queryStatUpdate(extime.Time(), [(u'beans', 2)])
    athena.expose(run)
